// sec_de3s.cpp
// code for sec_de3s.h
// copyright SafeTP Development Group, Inc., 2000  Terms of use are as specified in license.txt

#include <stdio.h>          // printf
#include <string.h>         // strlen

#include "sec_de3s.h"       // this module
#include "keyutils.h"       // LoadKey
#include "globrand.h"       // getRandomBytes
#include "exc.h"            // xsecurity
#include "warn.h"           // warning
#include "blokutil.h"       // Left, Right, ||, etc.
#include "sdsa.h"           // DSA*
#include "keydb.h"          // handleServerKey
#include "sockutil.h"       // formatAddress


// logging levels:
//   0  no logging
//   1  diagDataBlock1: keys and challenges
//   2  everything
#ifndef PROTOCOL_LOGGING
#  define PROTOCOL_LOGGING 0
#endif

#if PROTOCOL_LOGGING >= 1
#  define diagDataBlock1(block, label) block.print(label)
#else
#  define diagDataBlock1(block, label)
#endif

#if PROTOCOL_LOGGING >= 2
#  define diagnostic printf
#  define diagDataBlock(block, label) diagDataBlock1(block, label)
#  define diagDataBlock2(block) diagDataBlock1(block, #block)
#else
   static int dummyPrintf(char const *, ...) { return 0; }
#  define diagnostic dummyPrintf
#  define diagDataBlock(block, label)
#  define diagDataBlock2(block)
#endif

// temp
#if 0
#  define tdiagnostic printf
#  define tdiagDataBlock(block, label) block.print(label)
#else
#  define tdiagnostic dummyPrintf
#  define tdiagDataBlock(block, label)
#endif


static void loadKey(char const *name, DataBlock &block, bool forceprotected=false)
{
  string s(name);
  if (!LoadKey(s, block, forceprotected)) {
    
    xfailure(stringb("LoadKey for " << name << " failed"));
  }
}


// ------------------ DE3S_Shared ---------------------------------
DE3S_Shared::DE3S_Shared(bool ac)
  : sha1HMacSigner(NULL),
    sha1HMacVerifier(NULL),
    amClient(ac)
{}

DE3S_Shared::~DE3S_Shared()
{
  delete sha1HMacSigner;
  delete sha1HMacVerifier;
}


// =================== DE3S_Control ==========================
DE3S_Control::DE3S_Control(KeyEnvironment *env, IPAddress ma, IPAddress pa)
  : DE3S_Shared(env != NULL /*amClient*/),
    controlTripleDesEnc(NULL),
    controlTripleDesDec(NULL),
    myAddress(ma),
    peerAddress(pa),
    keyEnvironment(env)
{
  if (amClient()) {
    // load client keys
    loadKey("ElGamal/public.key", clientElGamalPublicKey);
    diagDataBlock1(clientElGamalPublicKey, "ElGamal public key");

    loadKey("ElGamal/private.key", clientElGamalPrivateKey);
    diagDataBlock1(clientElGamalPrivateKey, "ElGamal private key");
  }

  else {
    // load server keys
    loadKey("DSA/public.key", brandedServerDSAPublicKey, true);
      // need to forceProtected, in case we're using KEYS_IN_REGISTRY
    diagDataBlock1(brandedServerDSAPublicKey, "branded DSA public key");

    loadKey("DSA/private.key", serverDSAPrivateKey, true);
      // need to forceProtected, in case we're using KEYS_IN_REGISTRY
    diagDataBlock1(serverDSAPrivateKey, "DSA private key");

    // de-brand public key (also verifies the brand)
    DSABrandedPublicKey branded(brandedServerDSAPublicKey);
    branded.DSAPublicKey::encode(serverDSAPublicKey);
    diagDataBlock1(serverDSAPublicKey, "DSA public key");
  }

  state = S_SERVER_PUBLIC_KEY;
  sendSequenceNumber = INITIAL_SEQUENCE_NUMBER;
  receiveSequenceNumber = INITIAL_SEQUENCE_NUMBER;
}


DE3S_Control::~DE3S_Control()
{
  delete controlTripleDesEnc;
  delete controlTripleDesDec;
}


// ------------------- ADAT negotiation ----------------------
// this returns true if the (client? "client" : "server")
// is sending data in this state
bool DE3S_Control::isSending(bool client) const
{
  if (client == true) {
    // client
    return state==S_CLIENT_PUBLIC_KEY ||
           state==S_SERVER_CHALLENGE;
  }
  else {
    // server
    return state==S_SERVER_PUBLIC_KEY ||
           state==S_MASTER_KEY;
  }
}


bool DE3S_Control::hasOutgoingAdat() const
{
  return isSending(amClient());
}


STATICDEF DataBlock DE3S_Control::makeUpChallenge()
{
  return getRandomBytes(CHALLENGE_LENGTH);
}


// return a datablock containing a string, *without* the null terminator
static DataBlock tagBlock(char const *tag)
{
  return DataBlock(tag, strlen(tag));
}


DataBlock DE3S_Control::make3DESKey(
  char const *leftTag, char const *rightTag) const
{
  return
    Sha1(masterKey || tagBlock(leftTag)) ||
    Left(4, Sha1(masterKey || tagBlock(rightTag)));
}

DataBlock DE3S_Control::makeShaHMacKey(char const *tag) const
{
  return Sha1(masterKey || tagBlock(tag));
}

DataBlock DE3S_Control::make3DESIV(char const *tag) const
{
  return Left(8, Sha1(masterKey || tagBlock(tag)));
}


void DE3S_Control::makeExtraKeys()
{
  // compute new keys
  clientControl3DesKey = make3DESKey("client 3des control left",
                                     "client 3des control right");
  serverControl3DesKey = make3DESKey("server 3des control left",
                                     "server 3des control right");

  clientControl3DesIV = make3DESIV("client control 3des iv");
  serverControl3DesIV = make3DESIV("server control 3des iv");

  clientData3DesKey = make3DESKey("client 3des data left",
                                  "client 3des data right");
  serverData3DesKey = make3DESKey("server 3des data left",
                                  "server 3des data right");

  clientSha1HMacKey = makeShaHMacKey("client shahmac");
  serverSha1HMacKey = makeShaHMacKey("server shahmac");

  // display them for diagnostic purposes
  diagDataBlock1(clientControl3DesKey, "client control 3DES key");
  diagDataBlock1(serverControl3DesKey, "server control 3DES key");

  diagDataBlock1(clientData3DesKey, "client data 3DES key");
  diagDataBlock1(serverData3DesKey, "server data 3DES key");

  diagDataBlock1(clientSha1HMacKey, "client SHA1HMAC key");
  diagDataBlock1(serverSha1HMacKey, "server SHA1HMAC key");

  // make 3des objects
  controlTripleDesEnc = new CBC3DESEncTrans(
    amClient()? clientControl3DesKey : serverControl3DesKey,
    amClient()? clientControl3DesIV  : serverControl3DesIV);
  controlTripleDesDec = new CBC3DESDecTrans(
    amClient()? serverControl3DesKey : clientControl3DesKey,
    amClient()? serverControl3DesIV  : clientControl3DesIV);

  // make sha1hmac objects
  sha1HMacSigner = new SHA1HMAC(
    amClient()? clientSha1HMacKey : serverSha1HMacKey);
  sha1HMacVerifier = new SHA1HMAC(
    amClient()? serverSha1HMacKey : clientSha1HMacKey);
}


void DE3S_Control::getNextOutgoingAdat(DataBlock &block)
{
  if (amClient() == false) {     // server
    if (state == S_SERVER_PUBLIC_KEY) {
      block = brandedServerDSAPublicKey;
    }

    else {
      xassert(state == S_MASTER_KEY);

      // compute challenge
      serverChallenge = makeUpChallenge();
      diagDataBlock1(serverChallenge, "server challenge");

      // compute master key
      masterKey = getRandomBytes(MASTER_KEY_LENGTH);
      diagDataBlock1(masterKey, "master key");

      // compute final message
      block = elGamalEncrypt(clientElGamalPublicKey, masterKey) ||
              clientChallenge || serverChallenge;
      appendIPAddress(block, myAddress);
      dsaSign(serverDSAPrivateKey, block);

      // compute the extra keys
      makeExtraKeys();
    }
  }

  else {     // client
    if (state == S_CLIENT_PUBLIC_KEY) {
      // compute challenge
      clientChallenge = makeUpChallenge();
      diagDataBlock1(clientChallenge, "client challenge");

      // compute final message
      block = clientElGamalPublicKey || clientChallenge;
    }

    else {
      xassert(state == S_SERVER_CHALLENGE);

      // compute the main keys
      makeExtraKeys();

      // compute block to send
      block = serverChallenge;
      controlTripleDesEnc->trans(block);
    }
  }

  state = (State)(state + 1);
}


bool DE3S_Control::expectingIncomingAdat() const
{
  // I'm expecting data during precisely those states
  // where the other guy is sending data
  return isSending(!amClient());
}


void DE3S_Control::incomingAdat(DataBlock &block)
{
  if (amClient() == false) {     // server
    if (state == S_CLIENT_PUBLIC_KEY) {
      // pull apart client challenge and client ElGamal public key
      clientChallenge = removeBlock(block, CHALLENGE_LENGTH);
      diagDataBlock1(clientChallenge, "client challenge");

      clientElGamalPublicKey = block;
      diagDataBlock1(clientElGamalPublicKey, "ElGamal public key");
    }

    else {
      xassert(state == S_SERVER_CHALLENGE);

      // un-3des
      controlTripleDesDec->trans(block);
      diagDataBlock1(block, "client's challenge response");

      // compare to serverChallenge
      if (block != serverChallenge) {
        xsecurity("Client failed server's challenge.");
      }
      else {
        diagnostic("client passed server's challenge\n");
      }
    }
  }

  else {     // client
    if (state == S_SERVER_PUBLIC_KEY) {
      // extract key
      brandedServerDSAPublicKey = block;
      diagDataBlock1(brandedServerDSAPublicKey, "branded DSA public key");

      // de-brand key (and verify brand)
      DSABrandedPublicKey branded(brandedServerDSAPublicKey);
      branded.DSAPublicKey::encode(serverDSAPublicKey);
      diagDataBlock1(serverDSAPublicKey, "DSA public key");

      // consult public-key database
      xassert(keyEnvironment != NULL);
      handleServerKey(brandedServerDSAPublicKey,
                      peerAddress /*host ip*/,
                      *keyEnvironment);
    }

    else {
      xassert(state == S_MASTER_KEY);

      // check the DSA signature
      dsaVerify(serverDSAPublicKey, block,
                "Server's signature is invalid.");

      // pull apart concatenated pieces
      IPAddress serverIP = removeIPAddress(block);
      if (serverIP != peerAddress) {
        // have to make a policy decision; the called code can
        // throw xsecurity if it doesn't want to allow it
        keyEnvironment->policy.
          serverIPMismatch(serverIP, peerAddress, serverDSAPublicKey);
      }

      serverChallenge = removeBlock(block, CHALLENGE_LENGTH);
      diagDataBlock1(serverChallenge, "server challenge");

      DataBlock clientChallengeTry = removeBlock(block, CHALLENGE_LENGTH);
      diagDataBlock1(clientChallengeTry, "server's challenge response");

      if (clientChallengeTry != clientChallenge) {
        xsecurity("Server failed client's challenge.");
      }
      else {
        diagnostic("server passed client's challenge\n");
      }

      // what remains is the encrypted master key
      diagDataBlock1(block, "encrypted master key");
      masterKey = elGamalDecrypt(clientElGamalPrivateKey, block);
      diagDataBlock1(masterKey, "master key");
    }
  }

  state = (State)(state + 1);
}


// ---------------- Steady-state encoding --------------------------
void DE3S_Control::encode(DataBlock &data)
{
  int srclen = data.getDataLen();
  xassert(state == S_AUTHENTICATED);

  tdiagDataBlock(data, "plaintext");

  // append flag byte
  appendNBO32(data, FLAG_CONTROL |
                    (amClient()? FLAG_CLIENT : FLAG_SERVER));
  tdiagDataBlock(data, "with flags");

  // append sequence number
  appendNBO32(data, sendSequenceNumber++);
  tdiagDataBlock(data, "with seqnum");

  // append SHA1HMac code
  appendDigest(data, *sha1HMacSigner);
  tdiagDataBlock(data, "with signature");

  // triple-des the thing
  controlTripleDesEnc->trans(data);
  tdiagDataBlock(data, "ciphertext");

  xassert(data.getDataLen() <= maximumEncodedSize(srclen));
}


int DE3S_Control::maximumEncodedSize(int decodedLen) const
{
  int len = decodedLen;
  //tdiagnostic("%d: len = %d\n", __LINE__, len);

  len += 8;                      // flag, sequence #
  //tdiagnostic("%d: len = %d\n", __LINE__, len);

  len += sha1HMacSigner->DigestSize();
  //tdiagnostic("%d: len = %d\n", __LINE__, len);

  len = controlTripleDesEnc->maxOutputSize(len);
  //tdiagnostic("%d: len = %d\n", __LINE__, len);

  return len;
}


void DE3S_Control::decode(DataBlock &data)
{
  int srclen = data.getDataLen();
  xassert(state == S_AUTHENTICATED);

  tdiagDataBlock(data, "ciphertext");

  // triple-des decode
  controlTripleDesDec->trans(data);
  tdiagDataBlock(data, "un-3des'd");

  // check SHA1HMac code
  checkDigest(data, *sha1HMacVerifier,
              "SHA1HMac does not match.");
  tdiagDataBlock(data, "removed sig");

  // check sequence number
  checkNBO32(data, receiveSequenceNumber++,
             "Sequence number does not match.");
  tdiagDataBlock(data, "removed seqnum");

  // check flags
  checkNBO32(data, FLAG_CONTROL |
                   (amClient()? FLAG_SERVER : FLAG_CLIENT),
             "Flags word does not match.");
  tdiagDataBlock(data, "plaintext");

  xassert(data.getDataLen() <= maximumDecodedSize(srclen));
}


int DE3S_Control::maximumDecodedSize(int encodedLen) const
{
  int len = encodedLen;
  //tdiagnostic("%d: len = %d\n", __LINE__, len);

  len = controlTripleDesDec->maxOutputSize(len);
  //tdiagnostic("%d: len = %d\n", __LINE__, len);

  len -= sha1HMacVerifier->DigestSize();
  //tdiagnostic("%d: len = %d\n", __LINE__, len);

  len -= 8;     // flags, sequence #
  //tdiagnostic("%d: len = %d\n", __LINE__, len);

  return len;
}


// ---------------------- DES3_Data -------------------------
DE3S_Data::DE3S_Data(DE3S_Shared &sh)
  : shared(sh),
    dataTripleDesEnc(NULL),
    dataTripleDesDec(NULL),
    sequenceNumber(INITIAL_SEQUENCE_NUMBER),
    fileNumber(INITIAL_FILE_NUMBER),
    fileDSL(DSL_NONE)      // to detect failure to call newFile
{}

DE3S_Data::~DE3S_Data()
{
  deleteCryptoObjs();
}


void DE3S_Data::deleteCryptoObjs()
{
  delete dataTripleDesEnc;
  dataTripleDesEnc = NULL;

  delete dataTripleDesDec;
  dataTripleDesDec = NULL;
}


void DE3S_Data::buildCryptoObjs()
{
  // construct the initialization vector:
  // 64-bit big-endian representation of file #
  DataBlock IV;
  appendNBO32(IV, 0);           // high 32 bits are 0
  appendNBO32(IV, fileNumber);  // low 32 bits are file #

  // Some rationale on initialization vectors:  The basic idea is that we
  // only need, for each encrypted message (with a particular key), to use
  // a *different* IV.  We don't need secret IVs, nor random IVs.  The reason
  // secrecy is not necessary is, because of the construction, all IVs after
  // the first are transmitted in the clear anyway, as the ciphertext blocks
  // themselves.  The reason randomness isn't necessary is because we XOR the
  // IV with the first plaintext block, and *then* encrypt-- the encryption
  // effectively jumbles things up.  Even the first IV, which is 0, isn't a
  // problem because it will still be unique.
  //
  // Most of this reasoning is based on Schneier, pp. 193-194.
  // The CRC Handbook of Cryptography seems to confirm this.

  // Note that using the file # as the initialization vector means if the
  // client and server get out of sync, decryption becomes impossible and
  // the HMAC will not match.

  // construct crypto objects
  dataTripleDesEnc = new CBC3DESEncTrans(
    shared.amClient? shared.clientData3DesKey :
                     shared.serverData3DesKey,
    IV);

  dataTripleDesDec = new CBC3DESDecTrans(
    shared.amClient? shared.serverData3DesKey :
                     shared.clientData3DesKey,
    IV);
}


// I'm letting part of the object's initialization take place after
// the ctor because it's convenient to construct it early, but the
// data we need (the 3des keys) isn't available then
void DE3S_Data::ensureCryptoObjs() const
{
  // verify they're either both null, or both non-null
  xassert((dataTripleDesEnc == NULL) ==
          (dataTripleDesDec == NULL));

  // build them if necessary
  if (dataTripleDesEnc == NULL) {
    // this call is a violation of this fn being 'const'; however,
    // since 'ensureCryptoObjs' is called before any use of them,
    // the state change that occurs here is not visible from the
    // outside
    const_cast<DE3S_Data*>(this)->
      buildCryptoObjs();
  }
}


void DE3S_Data::encode(DataBlock &data)
{
  if (fileDSL == DSL_NONE) {
    xfailure("you have to call newFile before encode");
  }

  ensureCryptoObjs();

  int srclen = data.getDataLen();
  tdiagDataBlock(data, "plaintext");

  // flags
  appendNBO32(data, FLAG_DATA |
                    (shared.amClient? FLAG_CLIENT : FLAG_SERVER));
  tdiagDataBlock(data, "with flags block");

  // sequencing
  appendNBO32(data, fileNumber);
  appendNBO32(data, sequenceNumber++);
  tdiagDataBlock(data, "with sequencing info");

  // attach the message authentication code
  appendDigest(data, *shared.sha1HMacSigner);
  tdiagDataBlock(data, "signed");

  if (fileDSL == DSL_PRIVATE) {
    // encrypt it
    dataTripleDesEnc->trans(data);
    tdiagDataBlock(data, "cihpertext");
  }

  xassert(data.getDataLen() <= maximumEncodedSize(srclen, fileDSL));
}

int DE3S_Data::maximumEncodedSize(int decodedSize,
                                  DataSecurityLevel level) const
{
  ensureCryptoObjs();

  int len = decodedSize;
  len += 12;                  // flags, sequencing
  len += shared.sha1HMacSigner->DigestSize();
  if (level == DSL_PRIVATE) {
    len = dataTripleDesEnc->maxOutputSize(len);
  }
  return len;
}

int DE3S_Data::maximumEncodedSize(int decodedSize) const
{
  // largest encoding is with privacy on
  return maximumEncodedSize(decodedSize, DSL_PRIVATE);
}


void DE3S_Data::decode(DataBlock &data)
{
  if (fileDSL == DSL_NONE) {
    xfailure("you have to call newFile before decode");
  }

  ensureCryptoObjs();

  int srclen = data.getDataLen();
  tdiagDataBlock(data, "ciphertext");

  if (fileDSL == DSL_PRIVATE) {
    // decrypt
    dataTripleDesDec->trans(data);
    tdiagDataBlock(data, "decrypted");
  }

  // verify message authentication code
  checkDigest(data, *shared.sha1HMacVerifier,
              "SHA1HMac does not match.");
  tdiagDataBlock(data, "removed sig");

  // check sequence number
  checkNBO32(data, sequenceNumber++,
             "Sequence number does not match.");
  tdiagDataBlock(data, "removed seqnum");

  // check file #
  checkNBO32(data, fileNumber,
             "File number does not match.");
  tdiagDataBlock(data, "removed file #");

  // check flags
  checkNBO32(data, FLAG_DATA |
                   (shared.amClient? FLAG_SERVER : FLAG_CLIENT),
             "Flags word does not match.");
  tdiagDataBlock(data, "removed flags");

  if (!( data.getDataLen() <= maximumDecodedSize(srclen, fileDSL) )) {
    breaker();
  }

  xassert(data.getDataLen() <= maximumDecodedSize(srclen, fileDSL));
}


int DE3S_Data::maximumDecodedSize(int encodedSize,
                                  DataSecurityLevel level) const
{
  ensureCryptoObjs();

  int len = encodedSize;
  if (level == DSL_PRIVATE) {
    len = dataTripleDesDec->maxOutputSize(len);
  }
  len -= shared.sha1HMacVerifier->DigestSize();
  len -= 12;     // flags, sequencing
  return len;
}

int DE3S_Data::maximumDecodedSize(int encodedSize) const
{
  // max decoded happens when the data was *not* encrypted
  return maximumDecodedSize(encodedSize, DSL_INTEGRITY);
}


DataSecurityLevel DE3S_Data::getSupportedProtLevels() const
{        
  // this object does *not* support cleartext -- that is handled
  // by having the data simply never pass through the DE3S_Data
  // object
  return (DataSecurityLevel)(DSL_INTEGRITY | DSL_PRIVATE);
  
  // I continue to be annoyed by the erratic treatment enums gets
  // w/respect to bitwise or..
}


// called between data channel files
void DE3S_Data::newFile(DataSecurityLevel level)
{
  // check that level is among named levels
  DataSecurity::newFile(level);

  sequenceNumber = INITIAL_SEQUENCE_NUMBER;
  fileNumber++;
  fileDSL = level;

  // this will force them to be created anew
  deleteCryptoObjs();
}


char DE3S_Data::getCodeForLevel(DataSecurityLevel level) const
{
  switch (level) {
    case DSL_INTEGRITY:  return 'T';
    case DSL_PRIVATE:    return 'P';

    default:
      xfailure("invalid security level");
      return '!';        // silence warning
  }
}


DataSecurityLevel DE3S_Data::getLevelForCode(char code) const
{
  // Note: I am assuming the argument to PROT case-sensitive

  // see which level it is
  switch (code) {
    case 'S':    // 2228: "safe"
    case 'E':    // 2228: "confidential"
    case 'P':    // 2228: "private" (integrity + confidentiality)
      // unfortunately, my original implementation mapped S, E, and P
      // all to P, effectively.  so, rather than change the meaning of
      // existing letters, we continue to support the original meaning..
      return DSL_PRIVATE;

    case 'T':    // means (roughly) what 2228 means by "safe"
      // ..and instead *extend*, by using what was before an illegal code
      return DSL_INTEGRITY;

    default:
      // unknown code
      return DSL_NONE;
  }
}


// ---------------- DE3S_Provider -------------------------
DE3S_Provider::DE3S_Provider(KeyEnvironment *env, IPAddress myAddr,
                             IPAddress peerAddr)
{
  // I was forced to make these objects separate from the
  // provider by one of two things:
  //   1. g++ sucks
  //   2. I forgot to recompile after a certain change
  // since I already know that (1) is true I'm not going to
  // waste time finding out if it was responsible for this
  // particular issue
  _control = new DE3S_Control(env, myAddr, peerAddr);
  _data = new DE3S_Data(*_control);
}


DE3S_Provider::~DE3S_Provider()
{
  delete _control;
  delete _data;
}


// -------------- test code -----------------
// at this point, this just tests CBC3DES{Enc,Dec}Trans and SHA1HMAC
// (provider.cc gives everything a good workout)
#ifdef TEST_SEC_DE3S

#include "test.h"      // USUAL_MAIN

void entry()
{
  bool ok = true;
  DataBlock key("abcdefghijklmnopqrstuvwx", 24),
            IV("12345678", 8);

  CBC3DESEncTrans enc(key, IV);
  CBC3DESDecTrans dec(key, IV);
  TransPair desPair(enc, dec);
  ok &= desPair.test(100 /*iters*/);

  // the following lines don't compile, I don't
  // care why
  #if 0
  SHA1HMAC sign(key);
  SHA1HMAC verify(key);
  TransPair hmacPair(sign, verify);
  ok &= hmacPair.test(100 /*iters*/);
  #endif // 0

  PVAL(ok);
}


USUAL_MAIN

#endif // TEST_SEC_DE3S

