// sftpc.cc
// SafeTP command-line (stand-alone) client
// copyright SafeTP Development Group, Inc., 2000  Terms of use are as specified in license.txt

#include <iostream.h>      // cout
#include <stdlib.h>        // atoi, system
#include <stdio.h>         // FILE functions
#include <string.h>        // strstr
#include <ctype.h>         // tolower, isspace
#include <signal.h>        // signal

#include "provider.h"      // SecurityProvider
#include "exc.h"           // exceptions
#include "base64t.h"       // base64transform
#include "globrand.h"      // {read,save}RandomSeed
#include "sockutil.h"      // socket funcs, INVALID_SOCKET
#include "nonport.h"       // getCurrentUsername, readNonechoString
#include "filesrc.h"       // FileInputSource
#include "test.h"          // PVAL
#include "glob.h"          // glob
#include "strtokp.h"       // StrtokParse
#include "keyutils.h"      // sm_testKey
#include "syserr.h"        // xSysError

#include "addent.h"        // getEntropyFromConsole
#include "makekeys.h"      // GenerateElgamalKey
#include "selgamal.h"      // ElGamal_1024bit_parameters
#include "sdsa.h"          // DSABrandedPublicKey, knownDSAPublicKeyVersions

#include "sftpver.h"       // SFTP_version
#include "sftpcdoc.h"      // command documentation and aliases
#include "sftpc.h"         // this module


#ifndef NDEBUG
#  define diagnostic(expr) printDiagnostic(stringb(expr))
#else
#  define diagnostic(expr) ((void)0)
#endif

// true: CRLF separates text lines; false: LF separates text lines
#ifndef DEFAULT_LOCAL_CONVENTION
#  ifdef __UNIX__
#    define DEFAULT_LOCAL_CONVENTION false
#  else   // windows
#    define DEFAULT_LOCAL_CONVENTION true
#  endif
#endif

#ifdef _MSC_VER                                                         
  // this one has to do with comparing signed and unsigned
  #pragma warning(disable: 4018) // dob: I'm sick of seeing this warning
  
  // this one is about 'this' appearing in a member initializer list,
  // which I think is fine but MS apparently disagrees;
  // TODO: look at spec to see what they say about 'this' in member inits
  #pragma warning(disable: 4355) // dob: I'm sick of seeing this warning
#endif


// util
string getCurrentUsername()
{
  char buf[80];
  getCurrentUsername(buf, 80);
  return string(buf);
}

string readNonechoString(char const *prompt)
{
  char buf[80];
  readNonechoString(buf, 80, prompt);
  return string(buf);
}


// the initializers here should be maintained in the same
// order, and with the same spacing, as they appear in
// the class declaration
SFTPC::SFTPC()
  : security(NULL),
    securityName("X-SafeTP1"),

    keyDatabase(),
    keyEnvironment(*this /*policy*/, keyDatabase),

    serverPort(FTPD_PORT),
    serverName("localhost"),
    serverAddress(INADDR_NONE),

    control(NULL),

    dataBuffer(),

    PBSZ(0),
    digt(),
    sentFirstAuth(false),
    authenticated(false),
    dataSecLevel(DSL_PRIVATE),    // default level

    requestedPBSZ(0x8000),        // 32k

    transferPassively(true),      // 9/17/00: new default for ubiquitous firewalls
    asciiTransfers(false),
    localConventionIsCRLF(DEFAULT_LOCAL_CONVENTION),

    printOutgoing(false),
    interactivePrompting(false),
    quitProgram(false),
    printHashMarks(false),

    useRfc959(false),
    acceptNewKeysSilently(false),
    auto959Dropdown(false),
    alwaysAcceptServerIPMismatch(false),

    showDiagnostics(false),
    showAdats(false),
    useDebuggingLogin(false),
    quitAfterNegotiation(false),
    binaryTransferAnyway(false),
    localGlobbing(true)
{}

SFTPC::~SFTPC()
{
  if (security) {
    delete security;
  }
  if (control) {
    delete control;
  }
}


void SFTPC::printDiagnostic(char const *msg)
{
  if (showDiagnostics) {
    cerr << msg << endl;
  }
}


// read the next reply, and if it is encrypted, decrypt it; then return it
Reply SFTPC::readReply()
{
  // wait for the reply
  diagnostic("waiting for server reply");
  Reply reply(*control);

  // look at reply code
  ReplyCode code = reply.getCode();

  // decrypt the message if necessary
  if (first(code) == RCF_ENCRYPTED) {
    diagnostic((int)reply.getCode() << " " << reply.getAllTextAsOneLine());

    // it can only be encrypted if we're authenticated
    xassert(authenticated);

    // allocate and prepare block
    string encoded = reply.getAllTextAsOneLine();
      // todo: this is wrong for multiline responses...
    int encodedLen = encoded.length();
    int decodedLen = 1 +   // null at end
      security->control().maximumDecodedSize(
        base64decoder.maxOutputSize(encodedLen));
    decodedLen = mymax(decodedLen, encodedLen);
    DataBlock block(decodedLen);
    block.setFromString(encoded);

    // un-64
    base64decoder.trans(block);

    // decrypt
    security->control().decode(block);

    // add a null in case the reply didn't include
    // a final CRLF (allowed by RFC 2228)
    block.addNull();

    // parse it as a reply, now that it's decoded
    char const *temp = (char const*)block.getDataC();
    Reply decryptedReply(temp);

    // print it
    cout << decryptedReply.getAllText();    // already has CRLF

    return decryptedReply;
  }

  else {
    // print reply
    if (showAdats || !reply.containsADAT()) {
      cout << reply.getAllText();             // already has CRLF
    }

    // it can only be plaintext if control channel unencrypted
    xassert(!isControlChannelEncrypted());

    // doesn't need to be decrypted
    return reply;
  }
}


Reply SFTPC::readFinalReply()
{
  // loop while the reply is intermediate
  int intermediates = 0;
  for(;;) {
    // get and decrypt reply, then print it
    Reply reply = readReply();

    // do DIGT and ADAT processing
    handleAdatAndDigt(reply);

    // look at reply code
    if (first(reply.getCode()) != RCF_POSITIVE_PRELIMINARY) {
      return reply;
    }
    else {
      intermediates++;
      if (intermediates > 1) {
        cout << "warning: server has violated FTP protocol by\n"
             << "         sending " << intermediates
             << " intermediate replies (1 is max)\n";
      }
    }
  }
}


// we expect a reply; listen for the reply, decode it if
// necessary, display it, then keep listening IF the reply
// was a preliminary reply (otherwise return final code)
ReplyCode SFTPC::readFinalReplyCode()
{
  return readFinalReply().getCode();
}


// get the reply, throw exception on error reply code
ReplyCode SFTPC::readAndCheckFinalReplyCode()
{
  // read, print reply
  Reply reply = readFinalReply();

  // confirm is ok
  checkReplyCode(reply);

  // return the code itself
  return reply.getCode();
}


// throw xReply if there is a problem
void SFTPC::checkReplyCode(Reply const &reply) const
{
  ReplyCodeFirst f = first(reply.getCode());
  if (!( f == RCF_POSITIVE_COMPLETION ||
         f == RCF_POSITIVE_INTERMEDIATE )) {
    xReply x(reply);
    THROW(x);
  }

  // 6/7/99: I had been allowing RCF_POSITIVE_INTERMEDIATE, which
  //         doesn't appear to be a possible return from readFinalReply.
  // 6/8/99: doh!  it's "preliminary" which can't be returned;
  //         "intermediate" can be (and is; e.g., "user" response is 331)
}


void SFTPC::handleAdatAndDigt(Reply &reply)
{
  // the reply may go into the DIGT calculation
  if (isDigtActive()) {
    digt.add(reply);
  }

  // see if the reply contains some data
  bool replyHasAdat = reply.containsADAT();

  // process the ADAT if necessary
  if (replyHasAdat) {
    // make sure 'security' is expecting this
    xassert(security &&
            security->control().expectingIncomingAdat());

    // get text
    string text = reply.getAllTextAsOneLine();

    // un-64 the reply text
    DataBlock adat;
    base64decode(adat, text + Reply::ADATTagLen);

    // what did we get?
    if (showAdats) {
      adat.print("server adat");
    }

    // pass this to the security mechanism
    security->control().incomingAdat(adat);
  }
}


void SFTPC::sendRequestToWire(Request const &req)
{
  // send the request
  diagnostic("sending request: " << req.getText());
  req.send(control->socket);

  // possibly add it to the digest calculation
  if (isDigtActive()) {
    digt.add(req);
  }
}


// send a request; this request is *not* encrypted as it is passed
// don't process the reply; don't even take it from the socket
void SFTPC::sendRequest(Request const &req)
{
  if (printOutgoing) {
    // this is how "ftp" does it
    cout << "--> " << req.getTextNoPassword() << endl;
  }

  if (isControlChannelEncrypted()) {
    // allocate and prepare buffer
    string const &decoded = req.getText();
    int decodedLen = decoded.length();
    int encodedLen = 1 +       // null terminator
      base64encoder.maxOutputSize(
        security->control().maximumEncodedSize(decodedLen));
    DataBlock block((byte const*)decoded.pcharc(), decodedLen,
                    mymax(decodedLen, encodedLen));

    // encrypt
    security->control().encode(block);

    // base64
    base64encoder.trans(block);

    // construct Request to hold it
    Request encryptedReq(CMD_ENC,
                         (char const*)block.getDataC());

    // send it
    diagnostic("sending plaintext request: " << req.getTextNoPassword());
    sendRequestToWire(encryptedReq);
  }

  else {       // control channel is in the clear
    sendRequestToWire(req);
  }
}


// send the request and process the reply (expecting a success reply)
void SFTPC::checkedRequest(Request const &req)
{
  // send request
  sendRequest(req);

  // process reply
  readAndCheckFinalReplyCode();
}


// request variants
void SFTPC::sendRequest(RequestCommand cmd, char const *args)
{
  sendRequest(Request(cmd, args));
}

void SFTPC::checkedRequest(RequestCommand cmd, char const *args)
{
  checkedRequest(Request(cmd, args));
}


// manages all communication from AUTH to "23{4,5} Security data exchange complete."
bool SFTPC::negotiationPart2()
{
  // create the security provider
  security = SecurityProvider::
    findSecurityMechanism(securityName,
                          &keyEnvironment,
                          getLocalAddress(control->socket),
                          getRemoteAddress(control->socket));
  if (!security) {
    xfailure(stringb("unknown security mechanism: " << securityName));
  }

  cout << "Starting negotiation...\n";

  // send an auth request (will be added to digt automatically)
  sentFirstAuth = true;
  sendRequest(CMD_AUTH, securityName);

  // get reply, process ADAT if included
  ReplyCode authReplyCode = readFinalReplyCode();
  if (first(authReplyCode) != RCF_POSITIVE_COMPLETION &&
      first(authReplyCode) != RCF_POSITIVE_INTERMEDIATE) {
    // AUTH not understood
    delete security;
    security = NULL;
    return false;
  }

  // send an ADAT if necessary
  while (security->control().hasOutgoingAdat()) {
    // get data
    DataBlock adat;
    security->control().getNextOutgoingAdat(adat);

    // what are we sending?
    if (showAdats) {
      adat.print("client adat");
    }

    // encode as base64
    string b64adat = base64encode(adat);

    // send it (will be added to digt automatically)
    sendRequest(CMD_ADAT, b64adat);

    // get reply, process ADAT if included
    readAndCheckFinalReplyCode();
  }

  // if we don't have data to send, but are expecting some, that
  // is an error, because we have no way to 'prompt' the server
  // for more
  xassert(!security->control().expectingIncomingAdat());

  // from now on all control channel communication is encrypted
  authenticated = true;

  // compare digests
  compareDigests();

  cout << "Negotiation completed.\n";

  return true;
}


void SFTPC::compareDigests()
{
  // ask for server's digest
  sendRequest(CMD_DIGT);

  // get server's reply
  Reply reply = readFinalReply();

  #if 0   // old; changed it after talking with Dan on 9/1/99
    // check it (we require that the server support this command,
    // because if we allow it not to, then a hacker has an easy
    // way to defeat it)
    xassert(first(reply.getCode()) == RCF_POSITIVE_COMPLETION);
  #else
    // we will allow the server to not support DIGT; the rationale
    // here is that if a hacker can insert a bogus failure response,
    // he/she can just as easily insert a bogus DIGT value; that is,
    // we are using a threat model where the control channel is believed
    // *authentic* but not necessarily *confidential*, as the former
    // would require an on-line attack but the latter can utilize an
    // off-line attack (e.g., on a weak encryption algorithm)
    if (first(reply.getCode()) != RCF_POSITIVE_COMPLETION) {
      cout << "Server doesn't understand DIGT command.  Proceeding anyway.\n";
      return;
    }
  #endif

  // find where the base-64 digest string starts
  enum { DIGTEQLEN = 5 };    // length of "DIGT="
  string replyText = reply.getAllTextAsOneLine();
  char const *digteq = strstr(replyText, "DIGT=");
  xassert(digteq);           // otherwise the server didn't include a digest!

  // decode the base-64 data
  DataBlock serverDigest;
  base64decode(serverDigest, digteq + DIGTEQLEN);
  //int foobarbaz = replyText.length();  // ?
  //serverDigest.print("server's DIGT");

  // retrieve our own digest value
  DataBlock clientDigest = digt.getDigt();
  //clientDigest.print("client's DIGT");

  // compare
  if (serverDigest != clientDigest) {

    // This is the first of several similarly-formatted messages that
    // sftpc can generate.  The idea here is to blow the whistle rather
    // loudly, so an attacker considering mounting an attack that would
    // provoke such a message will decide it's too high-profile.  Of
    // course, for this to work, the false-positive rate must be low.

    cout << "\n"
            "         ************************\n"
            "         *   Digest Mismatch!   *\n"
            "         ************************\n"
            "\n"
            "The server saw a different negotiation sequence than I\n"
            "did.  This could be the result of someone tampering with\n"
            "the data while in transit.  You could just try again...\n"
            "If the problem persists, contact your network administrator;\n"
            "the situation must be resolved before communications with\n"
            "this server can be trusted.\n"
            "\n"
            ;

    xfailure("Connection aborted due to digest mismatch.");
  }
}


// implements dropdown behavior
bool SFTPC::negotiationPart1(int &retval)
{
  #define FAILRETURN(val) retval=val; return false /* user-; */

  // do security mechanism negotiation
  if (!useRfc959) {
    if (!negotiationPart2()) {
      // server responded negatively to AUTH command
      if (keyDatabase.containsKey(serverAddress)) {
        cout << "\n"
                "        ************************\n"
                "        *   Dropdown Attack!   *\n"
                "        ************************\n"
                "\n"
                "The server claims to not understand the AUTH mechanism,\n"
                "but we have a key on record for this server!  Possibilities:\n"
                "  - An attacker is attempting to fool sftpc into using an\n"
                "    insecure protocol.\n"
                "  - SafeTP has been turned off at the server.\n"
                "  - The prior contact was on another port; did you try 353?\n"
                "If you want to force sftpc to connect insecurely anyway, use\n"
                "the -9 switch.\n"
                ;
        FAILRETURN(5);
      }
      else if (auto959Dropdown) {
        cout << "Server doesn't understand AUTH mechanism.  Reverting\n"
                "to insecure protocol (because -Q was specified).\n";
        useRfc959 = true;
      }
      else {
        cout << "Exiting because server doesn't understand AUTH mechanism.\n"
                "There are several ways to proceed:\n"
                "  - Connect to the server using the normal (insecure) FTP\n"
                "    protocol.  To do this, give sftpc the -9 flag.\n"
                "    Be aware that your password might be eavesdropped!\n"
                "  - Use a completely different system, such as scp or Kerberos.\n"
                "  - Try a different port.  In particular, some sites have\n"
                "    SafeTP on port 353 instead of the default port (21):\n"
                "      % sftpc the.server 353\n"
                ;
        FAILRETURN(4);     // failed negotiation
      }
    }
    else {
      // after successful negotiation, do data channel protection,
      // if the user wants something other than cleartext
      if (dataSecLevel != DSL_CLEAR &&
          setDataEncryption(dataSecLevel)) {
        // successfully set prot level; we will assume the server's
        // printed response is adequate to inform user
      }
      else {
        dataSecLevel = DSL_CLEAR;
        printf("Data channel is currently unencrypted (change with 'prot').\n");
      }
    }
  }

  #undef FAILRETURN
  return true;     // success
}


// ------------------ data channel stuff ---------------------
// bind a listener socket, tell server about it, return it
SOCKET SFTPC::setupActiveConnection()
{
  // bind and listen to a socket
  SOCKET dataListener = listen_socket(PORT_ANY);

  // determine which port was bound
  sockaddr_in localPortAddr;
  getSockName(dataListener, localPortAddr);

  // determine the IP address by which the server knows us
  sockaddr_in localIPAddr;
  getSockName(control->socket, localIPAddr);

  // format a string to contain this information for
  // the PORT command
  long ip = ntohl(localIPAddr.sin_addr.s_addr);
  int port = ntohs(localPortAddr.sin_port);
  diagnostic("bound socket to local port " << port);

  // write the a,a,a,a,p,p string
  string commandText = constructPortArg(ip, port);

  // send the PORT command, process the reply
  checkedRequest(CMD_PORT, commandText);

  // return the dataListener socket
  return dataListener;
}


// ask the server to bind a listener, and connect to it, and
// return the connected socket (this relies on the server being
// able to accept() a connection even though the corresponding
// data transfer hasn't yet begun)
SOCKET SFTPC::setupPassiveConnection()
{
  // send PASV command
  sendRequest(CMD_PASV);

  // get reply
  Reply reply = readReply();
  checkReplyCode(reply);       // minor fix: w/o this, errors produce a confusing message

  // parse the address and port
  IPAddress addr;
  int port;
  parsePasvArgument(reply.getAllTextAsOneLine(),
                    addr, port);

  // connect to that address and port
  diagnostic("trying to connect to " << formatAddress(addr) <<
             ", port " << port);
  return connect_socket(addr, port);
}


// ask the server to switch to a specific file transfer mode
// (ascii or image (binary))
void SFTPC::requestAsciiMode(bool ascii)
{
  // debugging: to test my CRLF routines, I want to have the server send
  //            files in binary, but treat them as if they were sent as
  //            ascii
  if (binaryTransferAnyway) {
    ascii = false;
  }

  // send TYPE A (ascii) or TYPE I (image/binary),
  //   and verify the server is happy with that
  checkedRequest(CMD_TYPE, ascii? "A" : "I");

  // possible optimization: could remember what mode we think
  // the server is in, and only change it if it's different than
  // what we want; my usual ftp client does *not* do this
  // optimization, which makes me suspect some servers could be
  // flaky about remembering the transfer mode state...
}


// active: send PORT and return listening socket
// passive: send PASV and return connected data channel socket
SOCKET SFTPC::dataChannelSetup(RequestCommand cmd)
{
  // set transfer mode
  requestAsciiMode(transferModeForCmd(cmd));

  if (transferPassively) {
    // PASV and connect()
    return setupPassiveConnection();
  }
  else {
    // PORT and listen()
    return setupActiveConnection();
  }
}


bool SFTPC::transferModeForCmd(RequestCommand cmd) const
{
  if (cmd == CMD_LIST || cmd == CMD_NLST) {
    // always do directory tranfers in ASCII mode
    return true;
  }
  else {
    // file transfers are in mode of user's choosing
    return asciiTransfers;
  }
}


// second phase of establishing a data connection, given
// that dataChannelSetup did first part, and 'dataListener'
// is its return value
SOCKET SFTPC::openDataChannel(SOCKET dataListener)
{
  // get the data channel
  if (!transferPassively) {
    // accept the connection
    diagnostic("waiting for server to connect to my data port");
    SOCKET dataChannel = accept_socket(dataListener);
    diagnostic("server connected: " << dataChannel);

    // stop listening
    close_socket(dataListener);

    return dataChannel;
  }
  else {
    // the dataListener is actually the already-connected socket
    return dataListener;
  }
}


void SFTPC::check150Reply()
{
  diagnostic("waiting for the 125/150");
  Reply interReply = readReply();
  ReplyCode code = interReply.getCode();
  if (first(code) == RCF_POSITIVE_PRELIMINARY) {
    diagnostic("got the 125/150 as expected");
  }
  else if (first(code) == RCF_POSITIVE_COMPLETION ||
           first(code) == RCF_POSITIVE_INTERMEDIATE) {
    diagnostic("oops.. that was supposed to be a 125/150");
    printf("unexpected reply: %d %s\n", 
           code, interReply.getAllTextAsOneLine().pcharc());
    // try to keep going anyway
  }
  else {
    // since the reply is an error, its text will already have
    // been printed
    xReply x(interReply);
    THROW(x);
    // not reached
  }
}


SOCKET SFTPC::startDataTransfer(RequestCommand cmd, char const *arg)
{
  // transfer mode, then PORT or PASV
  SOCKET dataListener = dataChannelSetup(cmd);

  // protocol semantics are that the incrementing is associated
  // with the *attempt*, so just before we send the command, we
  // increment (if the send itself fails, then we've lost our
  // connection to the server anyway)
  if (security && isDataChannelEncrypted()) {
    security->data().newFile(dataSecLevel);
  }

  // send the primary request
  sendRequest(cmd, arg);
  check150Reply();

  // get channel from listener
  SOCKET dataChannel = openDataChannel(dataListener);

  // setup buffer
  if (!isDataChannelEncrypted() &&
      dataBuffer.getAllocated() < UNENCRYPTED_BUFSIZE) {
        // this was the site of a bug in 1.20, because I was not testing
        // the sizes; thus, turning encryption off and back on would
        // truncated dataBuffer, causing an exception fail on data xfer 
    dataBuffer.setAllocated(UNENCRYPTED_BUFSIZE);
  }

  return dataChannel;
}


// convenient way to ensure a socket gets closed, so the protocol
// stream remains viable
class SocketCloser {
  SOCKET s;
public:
  SocketCloser(SOCKET sk) : s(sk) {}
  ~SocketCloser();
};

SocketCloser::~SocketCloser()
{
  close_socket(s);
}


// do a data transfer
//   cmd              arg                    localFname
//   ------           -------------------    -----------------
//   retr             remote fname           local fname to write to
//   stor,stou,appe   remote fname           local fname to read from
//   list,nlst        optional dir to list   unused
void SFTPC::dataTransfer(RequestCommand cmd, char const *arg,
                         char const *localFname)
{
  // check local file first
  FILE *fp = NULL;
  if (cmd == CMD_RETR || cmd == CMD_STOR ||
      cmd == CMD_STOU || cmd == CMD_APPE) {
    // open the file
    xassert(localFname);
    fp = fopen(localFname, cmd==CMD_RETR? "wb" : "rb");
    if (!fp) {
      THROW(xBase(stringb("failed to open local file " << localFname)));
    }
  }

  // dataChannel scope
  {
    SOCKET dataChannel;
    try {
      // PORT, etc.
      dataChannel = startDataTransfer(cmd, arg);
    }
    catch (...) {   // common cause: server says file doesn't exist for RETR
      if (fp != NULL) {
        // Dan caught a bug here.. wasn't checking for fp == NULL..
        fclose(fp);     // may as well, since we need this before removal
      }
      if (cmd == CMD_RETR) {
        // we already deleted any existing file, and now we failed to get
        // it from the server, so it's 0-length; go ahead and remove it
        removeFile(localFname);
      }
      diagnostic("re-throwing xReply");
      throw;          // just relay the error condition itself
    }

    SocketCloser closer(dataChannel);

    // do transfer
    bool transferMode = transferModeForCmd(cmd);
    switch (cmd) {
      case CMD_RETR: {
        // copy data from server to client

        FileOutputDest destStream(fp);

        // copy data from the socket to the file until the
        // socket is closed
        int totlen;
        try {
	  if (!printHashMarks) {
	    totlen = copyFromSocketToStream(
	      dataChannel, destStream, transferMode);
	  }
	  else {
	    EchoOutputStream echo(destStream);
	    totlen = copyFromSocketToStream(
	      dataChannel, echo, transferMode);
	    cout << endl;
	  }
        }
        catch (xBase &x) {
          cout << "WARNING: file " << localFname << " is probably "
                  "incomplete.\n";
                  
          // something subtle is happening here -- if the data channel is
          // closed unexpectedly, an xSocket is thrown.. if the toplevel
          // loop sees an xSocket, it assumes it was caused by a *control*
          // connection closure.. so we catch it here, and re-throw it as
          // an xBase (since C++ exceptions are typed statically), which
          // the toplevel loop (counterintuitively) assumes is less
          // serious
          throw x;  
          
          // all this nonsense exposes a serious flaw in the way I've
          // designed my exception-handling here.. unfortunately, I am
          // unaware of any techniques for improving the design, since
          // it's very difficult to reason about...
        }

        diagnostic("received " << totlen <<
                   " bytes to local file " << localFname);

        fclose(fp);
        break;
      }

      case CMD_STOR:
      case CMD_STOU:
      case CMD_APPE: {
        // all the commands that copy data from the client to
        // the server

        FileInputSource sourceStream(fp);

        // copy data from the file to the socket
        int totlen;
        try {
	  if (!printHashMarks) {
	    totlen = copyFromStreamToSocket(
	      sourceStream, dataChannel, transferMode);
	  }
	  else {
	    EchoInputStream echo(sourceStream);
	    totlen = copyFromStreamToSocket(
	      echo, dataChannel, transferMode);
	    cout << endl;
	  }
        }
        catch (xBase &x) {
          cout << "WARNING: remote file " << arg << " is probably "
                  "incomplete.\n";
          throw x;     // see above
        }

        diagnostic("sent " << totlen <<
                   " bytes from local file " << localFname);

        fclose(fp);
        break;
      }

      case CMD_LIST:
      case CMD_NLST: {
        // server sends data that is displayed on client's console

        diagnostic("waiting for directory listing");

        // copy data from the socket to the terminal until the
        // socket is closed
        FileOutputDest destStream(stdout);
        int totlen = copyFromSocketToStream(dataChannel, destStream,
                                            transferMode);

        diagnostic("received " << totlen <<
                   " bytes as directory listing");

        break;
      }

      default:
        xfailure("dataTransfer called with bad ftp command");
    }

    // dataChannel closed by closer
  }

  // get and check final 226
  readAndCheckFinalReplyCode();
}


STATICDEF void SFTPC::CRLF_to_LF(
  StreamInputSource &src, StreamOutputDest &dest)
{
  CRLF_to_something(src, dest, ::CRLF_to_LF);
}

STATICDEF void SFTPC::CRLF_to_CRLF(
  StreamInputSource &src, StreamOutputDest &dest)
{
  CRLF_to_something(src, dest, ::CRLF_to_CRLF);
}

STATICDEF void SFTPC::CRLF_to_something(
  StreamInputSource &src, StreamOutputDest &dest,
  FromCRLFTranslator translator)
{
  int bareCRs=0, bareLFs=0;
  translator(src, dest, bareCRs, bareLFs);
  if (bareCRs != 0  ||  bareLFs != 0) {
    cout << "warning: Data had " << bareCRs << " bare CRs and "
                                 << bareLFs << " bare LFs.\n"
         << "         (Confusing?  Try \"help crlf\".)\n";
  }
}

STATICDEF void SFTPC::LF_to_CRLF(
  StreamInputSource &src, StreamOutputDest &dest)
{
  int CRs=0;
  ::LF_to_CRLF(src, dest, CRs);
  if (CRs != 0) {
    cout << "warning: Data had " << CRs << " CRs.\n"
         << "         (Confusing?  Try \"help crlf\".)\n";
  }
}

SFTPC::CRLFTranslator SFTPC::getCRLFTranslator(bool fromCRLF)
{
  if (localConventionIsCRLF) {
    diagnostic("using CRLF_to_CRLF");
    return &CRLF_to_CRLF;
  }
  else {
    if (fromCRLF) {
      diagnostic("using CRLF_to_LF");
      return &CRLF_to_LF;
    }
    else {
      diagnostic("using LF_to_CRLF");
      return &LF_to_CRLF;
    }
  }
}


// copy data from a socket to a file
// handles ascii and binary transfers
// returns number of bytes received over network
int SFTPC::copyFromSocketToStream(
  SOCKET sourceSocket, StreamOutputDest &destStream, bool asciiMode)
{
  if (asciiMode) {
    // must convert from CRLF to native

    // funky source
    int totlen=0;
    EncryptedInputStreamer sourceStream(*this, sourceSocket, totlen);

    // call Flex-generated scanner to do translation
    CRLFTranslator translator = getCRLFTranslator(true /*fromCRLF*/);
    translator(sourceStream, destStream);

    return totlen;
  }

  else {
    // straight copying
    int totlen=0;
    for(;;) {
      int len = readEitherWay(sourceSocket);
      if (len == 0) {
        // end of file
        break;
      }

      destStream.write((char const*)dataBuffer.getDataC(), len);
      totlen+=len;
    }
    return totlen;
  }
}


// copy data from a file to a socket
// handles ascii and binary transfers
// returns number of bytes sent over network
int SFTPC::copyFromStreamToSocket(
  StreamInputSource &sourceStream, SOCKET destSocket, bool asciiMode)
{
  // calculate how much cleartext we can get
  int cleartextLen;
  if (isDataChannelEncrypted()) {
    cleartextLen = security->
      data().maximumCleartextSizeForBlock(PBSZ);
  }
  else {
    cleartextLen = UNENCRYPTED_BUFSIZE;
  }
  diagnostic("cleartextLen is " << cleartextLen);

  if (asciiMode) {
    // must convert from native to CRLF

    // funky dest
    int totlen=0;
    {
      // inside braces to be sure of dtor order, since the dtor does
      // protocol stuff
      EncryptedOutputStreamer destStream(
        *this, destSocket, cleartextLen, totlen);

      // call Flex-generated scanner to do translation
      CRLFTranslator translator = getCRLFTranslator(false /*fromCRLF*/);
      translator(sourceStream, destStream);
    }

    return totlen;
  }

  else {
    // straight copying

    // copy data from the file to the socket
    int totlen=0;
    for(;;) {
      // read from file
      xassert(cleartextLen <= dataBuffer.getAllocated());
      int len = sourceStream.read((char*)dataBuffer.getData(), cleartextLen);
      dataBuffer.setDataLen(len);

      // write to socket
      writeEitherWay(destSocket);
      totlen += len;

      if (len == 0) {   // eof
        break;
      }
    }

    return totlen;
  }
}


// read data from s into dataBuffer
void SFTPC::readCleartext(SOCKET socket)
{
  recvAllToEOFBlock(socket, dataBuffer,
                    dataBuffer.getAllocated());
}


// write data from dataBuffer to s
void SFTPC::writeCleartext(SOCKET socket)
{
  sendAll(socket, (char const*)dataBuffer.getDataC(),
                  dataBuffer.getDataLen());
}


// read data from 'socket', decrypt it into dataBuffer
void SFTPC::readAndDecrypt(SOCKET socket)
{
  // read block size
  long blockSize = recvNBO32(socket);
  xassert(blockSize <= PBSZ);

  // read that much data
  recvAllBlock(socket, dataBuffer, blockSize);

  // decrypt data
  xassert(security);
  security->data().decode(dataBuffer);
}


// encrypt the data in dataBuffer, send it to 'socket'
void SFTPC::encryptAndWrite(SOCKET socket)
{
  // encrypt data
  xassert(security);
  security->data().encode(dataBuffer);

  // send block size
  long blockSize = dataBuffer.getDataLen();
  xassert(blockSize <= PBSZ);
  sendNBO32(socket, blockSize);

  // send the data
  sendAll(socket, (char const*)dataBuffer.getDataC(), blockSize);
}


// auto-multiplexing
int SFTPC::readEitherWay(SOCKET socket)
{
  if (isDataChannelEncrypted()) {
    readAndDecrypt(socket);
  }
  else {
    readCleartext(socket);
  }
  return dataBuffer.getDataLen();
}

void SFTPC::writeEitherWay(SOCKET socket)
{
  if (isDataChannelEncrypted()) {
    encryptAndWrite(socket);
  }
  else {
    writeCleartext(socket);
  }
}


// ---------------- multiple-file commands ----------------------
void SFTPC::mlsCommand(char const *pattern)
{
  // get strings that match the pattern
  Queue<string> queue = getRemoteNames(pattern);

  // process each file in the queue
  while (!queue.isEmpty()) {
    // grab name
    string name = queue.dequeue();

    // print it
    cout << name << endl;
  }
}


void SFTPC::mgetCommand(char const *pattern)
{
  // get strings that match the pattern
  Queue<string> queue = getRemoteNames(pattern);

  // process them
  multipleFileCommand(queue, CMD_RETR);
}


void SFTPC::mputCommand(char const *pattern)
{
  // get strings that match the pattern
  Queue<string> queue = getLocalNames(pattern);

  // process them
  multipleFileCommand(queue, CMD_STOR);
}


void SFTPC::mdeleteCommand(char const *pattern)
{
  // get strings that match the pattern
  Queue<string> queue = getRemoteNames(pattern);

  // process them
  multipleFileCommand(queue, CMD_DELE);
}


// get a single-character response
char SFTPC::getOneChar(char const *prompt, char const *allowable)
{
  // write prompt
  cout << prompt;
  cout.flush();

  // read single character
  setRawMode(true);      // terminal buffering off
  for(;;) {
    char ch = getConsoleChar();
    if (strchr(allowable, ch)) {
      // ok
      setRawMode(false);
      cout << ch << endl;    // echo what we accepted
      return ch;
    }
    if (ch == 3) {           // in raw mode, I *do* see ^C, at least on some systems
      setRawMode(false);
      xfailure("user typed ^C");    // if I see it, bail
    }

    // beeping the terminal might be an option here, but whatever
  }
}


void SFTPC::multipleFileCommand(Queue<string> &queue, RequestCommand cmd)
{
  // process each file in the queue
  bool localOverride = false;
  while (!queue.isEmpty()) {
    // grab name
    string name = queue.dequeue();

    // print it
    cout << name;

    // prompt (if user wants it)
    if (interactivePrompting && !localOverride) {
      char ch = getOneChar("   ((y)es/(n)o/yes to (a)ll/(q)uit)? ",
                           "ynaq");
      if (ch == 'n') {
        continue;       // go to next name
      }
      if (ch == 'q') {
        return;         // stop altogether
      }
      if (ch == 'a') {
        localOverride = true;     // and proceed with this transfer
      }
    }
    else {
      cout << endl;
    }

    // do the data transfer
    for (;;) {     // poor-man's goto label
      try {
        if (cmd != CMD_DELE) {
          dataTransfer(cmd, name, name /*local name, if needed*/);
        }
        else {
          checkedRequest(cmd, name);
        }
      }
      catch (xReply &) {
        // the message will already have been printed, so just prompt
        // (can't use a goto here because of initialization of 'x' stuff)
        switch (getOneChar("(c)ontinue, (a)bort, or (r)etry? ", "car")) {
          case 'a':
            return;    // bail completely
          case 'r':
            continue;  // go to top of loop
        }
      }
      catch (xBase &x) {
        // the justification for all this mechanism?  when we are using
        // remote-side globbing, we sometimes get names like "a/b"; this
        // can be useful, but if the directory "a" doesn't exist, it will
        // fail; so this loop lets the user go make the directory, then
        // proceed
        cout << x << endl;
        switch (getOneChar("(c)ontinue, (a)bort, or (r)etry? ", "car")) {
          case 'a':
            return;    // bail completely
          case 'r':
            continue;  // go to top of loop
        }
      }

      break;
    }
  }
}


Queue<string> SFTPC::getRemoteNames(char const *pattern)
{
  // PORT, etc.
  SOCKET dataChannel =
    startDataTransfer(CMD_NLST, localGlobbing? (char const*)NULL : pattern);

  diagnostic("waiting for directory listing");

  // line parser
  int networkBytes = 0;
  EncryptedInputStreamer stream(*this, dataChannel, networkBytes);
  StreamLineReader liner(stream);

  // add all the strings to the queue
  Queue<string> queue;
  string s;
  while (liner.getNextLine(s)) {
    //diagnostic("enqueueing: " << s);
    queue.enqueue(s);
  }

  diagnostic("received entire directory listing (" << queue.count() <<
             " names, " << networkBytes << " bytes)");

  // close socket, and check 226 reply
  close_socket(dataChannel);
  readAndCheckFinalReplyCode();

  // filter based on pattern (if local globbing)
  if (localGlobbing) {
    filterStrings(queue, pattern);

    // quick feedback
    cout << queue.count() << " names match the pattern\n";
  }
  else {
    cout << queue.count() << " names returned\n";
  }

  return queue;
}


void SFTPC::filterStrings(Queue<string> &queue, char const *pattern)
{
  // move queue elements into source queue
  Queue<string> src(queue);
  queue.empty();

  // process them in order
  while (!src.isEmpty()) {
    string name = src.dequeue();

    if (glob(name, pattern)) {
      queue.enqueue(name);
    }
  }
}


STATICDEF bool SFTPC::getLocalNamesHelper(
  char const *fname, void *extra)
{
  Queue<string> *q = (Queue<string>*)extra;
  if (fileOrDirectoryExists(fname) &&
        // we need this check because some programs (emacs..) create symlinks
        // that don't link anywhere valid; calling isDirectory on such a file
        // will fail (and throw an exception, since sftpc installs an
        // exception-throwing 'fail' routine); but fileOrDirectoryExists simply
        // returns false when stat(2) fails, so this safely avoids the
        // exception (could also just catch the exception.. I'm undecided on
        // which is best)
      !isDirectory(fname)) {
    q->enqueue(fname);
  }
  return true;    // continue;
}

Queue<string> SFTPC::getLocalNames(char const *pattern)
{
  // retrieve all local names
  Queue<string> ret;
  applyToCwdContents(getLocalNamesHelper, &ret);

  // filter based on pattern
  filterStrings(ret, pattern);
  cout << ret.count() << " file names match the pattern\n";

  return ret;
}


// ------------- automatic online testing ------------------
// test binary-mode transfer
void SFTPC::testBinaryTransfers()
{
  // binary transfer mode
  TransferState bin(false /*ascii*/, false /*localcrlf*/,
                    false /*binaryAnyway*/);

  // small but pathological
  static char const smal[] =
    "\0abc\n\r\r\n\xff\xfe\x01\x02\0\tyadda smacker\0\0\0foo";
  DataBlock smallBlock(smal, sizeof(smal)-1);
  testSendReceiveBlocks(smallBlock, bin, smallBlock, bin);

  // medium and large sizes, simple structure
  int sizes[2] = { 2345, 480000 };
  loopi(TABLESIZE(sizes)) {
    // construct data
    int sz = sizes[i];
    DataBlock block(sz);
    loopj(sz) {
      block.getData()[j] = (byte)j;
    }
    block.setDataLen(sz);

    // send/recv
    testSendReceiveBlocks(block, bin, block, bin);
  }
}


// test text-mode transfer
void SFTPC::testTextTransfers()
{
  /*
   *  The idea here is I have a bunch of text files with
   *  various line ending conditions.  I want to force sftpc
   *  to do every kind of text transformation it knows how to
   *  do, and verify I get what I should.
   *
   *  The purpose here is *not* to test the *server*'s ability
   *  to do text transformations, so the binaryAnyway flag is
   *  always set whenever we tell sftpc to do an ascii transfer.
   *
   *  These tests will print warnings about malformed text;
   *  I neither check nor suppress this output.
   */

  // sizeof returns the length *including* the automatically-
  // appended null terminator, which I don't want included
  #define BLK(arr) DataBlock b_##arr(arr, sizeof(arr)-1)

  // well-structured CRLF text
  char const crlf1[] =
    "This is a text file.\r\n"
    "Its lines are separated by CRLF.\r\n"
    "This is the third and final line.\r\n";
  BLK(crlf1);

  // badly-structured CRLF text
  char const crlf2[] =
    "This is another file with CRLF separators.\r\n"
    "Here (\r) is an embedded CR.\r\n"
    "Here (\n) is an embedded LF.\r\n"
    "And finally, a line missing any separator.";
  BLK(crlf2);

  // crlf1 -> CRLF_to_LF -> lf1
  char const lf1[] =
    "This is a text file.\n"
    "Its lines are separated by CRLF.\n"
    "This is the third and final line.\n";
  BLK(lf1);

  // crlf2 -> CRLF_to_LF -> lf2
  char const lf2[] =
    "This is another file with CRLF separators.\n"
    "Here () is an embedded CR.\n"
    "Here () is an embedded LF.\n"
    "And finally, a line missing any separator.";
  BLK(lf2);

  // lf1 -> LF_to_CRLF -> crlf1 (correct inverse)

  // lf2 -> LF_to_CRLF -> crlf2b
  char const crlf2b[] =
    "This is another file with CRLF separators.\r\n"
    "Here () is an embedded CR.\r\n"
    "Here () is an embedded LF.\r\n"
    "And finally, a line missing any separator.";
  BLK(crlf2b);

  // crlf2 -> LF_to_CRLF -> crlf2c
  char const crlf2c[] =
    "This is another file with CRLF separators.\r\n"
    "Here () is an embedded CR.\r\n"
    "Here (\r\n) is an embedded LF.\r\n"
    "And finally, a line missing any separator.";
  BLK(crlf2c);

  // crlf2 -> CRLF_to_CRLF -> crlf2d
  char const crlf2d[] =
    "This is another file with CRLF separators.\r\n"
    "Here () is an embedded CR.\r\n"
    "Here () is an embedded LF.\r\n"
    "And finally, a line missing any separator.";
  BLK(crlf2d);
  #undef BLK

  // here are the set of inputs, transforms, and outputs
  // I believe to be correct
  enum TransType { CRLF2LF, LF2CRLF, CRLF2CRLF };
  struct {
    DataBlock const *start;    // can't use & because of something about ctors
    TransType tt;
    DataBlock const *end;
  } arr[] = {
    { &b_crlf1, CRLF2LF, &b_lf1 },
    { &b_crlf2, CRLF2LF, &b_lf2 },

    { &b_lf1, LF2CRLF, &b_crlf1 },
    { &b_lf2, LF2CRLF, &b_crlf2b },
    { &b_crlf2, LF2CRLF, &b_crlf2c },

    { &b_crlf2, CRLF2CRLF, &b_crlf2d }
  };

  // this loop runs each line in the above array
  loopi(TABLESIZE(arr)) {
    switch (arr[i].tt) {
      case LF2CRLF:
        testSendReceiveBlocks(     // ascii  local  binaryAnyway
          *arr[i].start, TransferState(true,  false, true),
          *arr[i].end,   TransferState(false, false, false));
        break;

      case CRLF2LF:
        testSendReceiveBlocks(     // ascii  local  binaryAnyway
          *arr[i].start, TransferState(false, false, false),
          *arr[i].end,   TransferState(true,  false, true));
        break;

      case CRLF2CRLF:
        testSendReceiveBlocks(     // ascii  local  binaryAnyway
          *arr[i].start, TransferState(true,  true,  true),
          *arr[i].end,   TransferState(false, false, false));
        break;
    }
  }
}


// send and receive some data, and compare what we get
// to what we expect
void SFTPC::testSendReceiveBlocks(
  DataBlock const &start, TransferState const &sendState,
  DataBlock const &end, TransferState const &receiveState)
{
  // I had been using the same name for the local and remote file
  // names, but that created problems when the local and remote
  // directories were the same

  // write the initial data to a temporary file
  start.writeToFile("tempfile.local");

  // transfer this file to the server
  setTransferState(sendState);
  userCommand("put tempfile.local tempfile.remote");

  // retrieve the file from the server
  setTransferState(receiveState);
  userCommand("get tempfile.remote tempfile.local");

  // read what we got into a buffer
  DataBlock got;
  got.readFromFile("tempfile.local");

  // compare what we got to what we expect
  if (end != got) {
    if (start.getDataLen() +
        end.getDataLen() +
        got.getDataLen() < 3000) {
      // print the failing data if it's not too big
      start.print("initial");
      end.print("expected");
      got.print("actually got");
    }
    else {
      // we already have 'got' as tempfile.local ; write the
      // others to disk, for later human inspection
      start.writeToFile("tempfile.start");
      end.writeToFile("tempfile.end");
    }

    xfailure("put/get failed to preserve data");
  }

  else {
    printf("   (files are identical)   \n");

    // if it worked, go ahead and delete the temp files
    userCommand("rm tempfile.remote");    // from server
    removeFile("tempfile.local");         // from cwd
  }
}


void SFTPC::setTransferState(TransferState const &state)
{
  #define C(v) v = state.v
  C(asciiTransfers);
  C(localConventionIsCRLF);
  C(binaryTransferAnyway);
  #undef C
}


// a series of automated tests to verify that various functions
// in sftpc and sftpd are working correctly
void SFTPC::runTests(bool passive)
{
  transferPassively = passive;

  // test a simple ls -l command
  userCommand("dir");

  // test failed send
  bool eatIt = true;
  try {
    userCommand("put crazy.nonexist.file");
    eatIt = false;
    xfailure("failed to fail to put crazy.nonexist.file");
  }
  catch (...) {
    if (!eatIt) {
      throw;
    }
  }

  // test real data transfers
  testBinaryTransfers();

  // interpose a failed receive
  eatIt = true;
  try {
    userCommand("get crazier.nonexist.file");
    eatIt = false;
    xfailure("failed to fail to get crazy.nonexist.file");
  }
  catch (...) {
    if (!eatIt) {
      throw;
    }
  }

  // bunch of text transfers
  testTextTransfers();

  // test multiple-file commands
  testMultipleFileCmds();
}


static void verifyFile(DataBlock const &blk, char const *fname)
{
  DataBlock temp;
  temp.readFromFile(fname);
  xassert(blk == temp);
}


// primary intent here is to test the code that selects
// file names and initiates transfers, rather than that
// each file's contents are transferred correctly
void SFTPC::testMultipleFileCmds()
{
  // some data to toss around
  DataBlock a("block A");
  DataBlock b("block B");
  DataBlock c("block C");

  // write them to files
  a.writeToFile("testmf.a");
  b.writeToFile("testmf.b");
  c.writeToFile("testmf.c");

  // send them
  interactivePrompting = false;
  asciiTransfers = false;
  userCommand("mput testmf.*");

  // just to verify it doesn't crash...
  userCommand("mls testmf.*");

  // delete the local versions
  removeFile("testmf.a");
  removeFile("testmf.b");
  removeFile("testmf.c");

  // get them back from the server
  userCommand("mget testmf.*");

  // verify they all arrived correctly
  verifyFile(a, "testmf.a");
  verifyFile(b, "testmf.b");
  verifyFile(c, "testmf.c");

  // remove them from server
  userCommand("mdelete testmf.*");

  // and locally
  removeFile("testmf.a");
  removeFile("testmf.b");
  removeFile("testmf.c");
}


// -------------- command line interpretation -----------------
// little helper
static string getCurDir()
{
  char buf[200];
  if (!getCurrentDirectory(buf, 200)) {
    return string("(error retrieving current directory!)");
  }
  else {
    return string(buf);
  }
}


void SFTPC::userCommand(char const *command)
{
  // parse the command into words
  StrtokParse tok(command, " \t");
  if (tok == 0) {
    return;       // user didn't type anything
  }

  // macros for querying command
  // (scoped by #undefining at end of fn)
  #define WORDnIS(str,n) (0==strcmp(tok[n], (str)))
  #define CMDIS(str) WORDnIS(str,0)
  #define CMD(str) else if (CMDIS(str))
  #define NO_ARG1 if (tok <= 1)
  #define ARG1IS(str) WORDnIS(str,1)
  #define ARG1(str) else if (ARG1IS(str))

  // setting variables conveniently
  #define SET_VAR(var, newval)     \
    var = newval;                  \
    cout << #var " is now " << var << endl /*user supplies semicolon*/

  #define TOGGLE_VAR(var) SET_VAR(var, !var)

  // toggle if no arguments, or set to match on/off arg
  #define TOGGLE_ON_OFF_VAR(var)  \
    NO_ARG1 {                     \
      TOGGLE_VAR(var);            \
    }                             \
    ARG1("on") {                  \
      SET_VAR(var, true);         \
    }                             \
    ARG1("off") {                 \
      SET_VAR(var, false);        \
    }   /* no semicolon! */

  // command constraints
  #define REQUIRE_ARGS(n)                                           \
    if (tok <= n) {                                                 \
      cout << "The " << tok[0] << " command requires at least "     \
           << n << " arguments.\n"                                  \
           << "Try \"help " << tok[0] << "\"\n";                    \
      return;                                                       \
    }

  #define ELSE_BAD_ARG1                             \
    else {                                          \
      cout << "Unknown argument to " << tok[0]      \
           << ": " << tok[1]                        \
           << "\nTry \"help " << tok[0] << "\"\n";  \
    }

  // convenient forms of arguments
  char const *arg1 = (tok<=1? NULL : tok[1]);
  char const *arg2 = (tok<=2? NULL : tok[2]);

  // convenient forms of text following some argument
  char const *afterCmd = (tok<=1? "" : command + tok.offset(1));
  //char const *afterArg1 = (tok<=2? "" : command + tok.offset(2));


  // the alias table that used to be here is now in sftpcdoc.cpp


  // -------------- non-protocol actions --------------
  if (CMDIS("debug")) {
    TOGGLE_ON_OFF_VAR(printOutgoing)
    ARG1("1") {
      TOGGLE_VAR(showDiagnostics);
    }
    ARG1("breaker") {
      breaker();
    }
    ARG1("dump") {
      dumpVariables();
    }
    ARG1("binaryAnyway") {
      TOGGLE_VAR(binaryTransferAnyway);
    }
    ARG1("localCRLF") {
      TOGGLE_VAR(localConventionIsCRLF);
    }
    ARG1("localGlobbing") {
      TOGGLE_VAR(localGlobbing);
    }
    ELSE_BAD_ARG1
  }
  CMD("passive") {
    TOGGLE_ON_OFF_VAR(transferPassively)
    ELSE_BAD_ARG1
  }
  CMD("ascii") {
    SET_VAR(asciiTransfers, true);
  }
  CMD("binary") {
    SET_VAR(asciiTransfers, false);
  }
  CMD("type") {
    REQUIRE_ARGS(1)
    SET_VAR(asciiTransfers, ARG1IS("a"));
  }
  CMD("lcd") {
    char const *destDir = arg1;
    if (!destDir) {
      // with no arguments, change to home directory
      destDir = getenv("HOME");
      if (!destDir) {
        cout << "Can't change to home directory because environment\n"
                "variable HOME is not defined.\n";
        goto endOfLcd;
      }
    }

    if (!changeDirectory(destDir)) {
      cout << "Couldn't change to " << destDir << endl;
    }
    else {
      cout << "Current local directory changed to " << getCurDir() << endl;
    }
    endOfLcd:
      ;       // gnu requires this (grumble grumble..)
  }
  CMD("lpwd") {
    cout << "Current local directory is " << getCurDir() << endl;
  }
  CMD("prompt") {
    TOGGLE_ON_OFF_VAR(interactivePrompting)
    ELSE_BAD_ARG1
  }
  CMD("quit") {
    quitProgram = true;
  }
  else if (command[0] == '!') {
    // advance to next non-ws after '!'
    char const *p = command+1;
    while (isspace(*p)) {
      p++;
    }
    if (system(p) == -1) {
      xsyserror("system");
    }
  }
  CMD("hash") {
    TOGGLE_ON_OFF_VAR(printHashMarks)
    ELSE_BAD_ARG1
  }

  // ------------- special protocol actions --------------
  CMD("sync") {
    if (emptyResponseQueue() == 0) {
      cout << "The response queue appears to be empty.\n";
    }
  }
  CMD("test") {
    NO_ARG1 {
      // run all tests in all modes
      static DataSecurityLevel const map[3] =
        { DSL_CLEAR, DSL_INTEGRITY, DSL_PRIVATE };
      for (int lvl=0; lvl<3; lvl++) {
        for (int psv=0; psv<2; psv++) {
          if (setDataEncryption(map[lvl])) {
            // only run the tests if the server supports the mode
            runTests(psv==1 /*passive*/);
          }
        }
      }
    }
    ARG1("active") {
      runTests(false /*passive*/);
    }
    ARG1("passive") {
      runTests(true /*passive*/);
    }
    ARG1("text") {
      testTextTransfers();
    }
    ARG1("binary") {
      testBinaryTransfers();
    }
    ARG1("multi") {
      testMultipleFileCmds();
    }
    ELSE_BAD_ARG1
  }
  CMD("quote") {
    checkedRequest(Request(afterCmd));
  }
  CMD("prot") {
    NO_ARG1 {
      cout << "current data protection level: " 
           << getDSLString(dataSecLevel) << endl;
    }
    ARG1("p") {
      setDataEncryption(DSL_PRIVATE);
    }
    ARG1("c") {
      setDataEncryption(DSL_CLEAR);
    }
    ARG1("t") {      // deliberately avoiding 's' because I botched its meaning earlier
      setDataEncryption(DSL_INTEGRITY);
    }
    ELSE_BAD_ARG1
  }

  // -------- "ftp"-like protocol actions ------------
  CMD("pwd") {
    checkedRequest(CMD_PWD);
  }
  CMD("cd") {
    if (tok >= 1) {
      checkedRequest(CMD_CWD, command + tok.offset(1));    // issues with spaces..
    }
    else {
      checkedRequest(CMD_CWD);       // cd to home directory
    }
  }
  CMD("cdup") {
    checkedRequest(CMD_CDUP);
  }
  CMD("dir") {
    dataTransfer(CMD_LIST, arg1, NULL);
  }
  CMD("nlist") {
    dataTransfer(CMD_NLST, arg1, NULL);
  }
  CMD("get") {
    REQUIRE_ARGS(1);
    dataTransfer(CMD_RETR, arg1, arg2? arg2 : arg1);
  }
  CMD("put") {
    REQUIRE_ARGS(1);
    if (arg2 != NULL) {
      dataTransfer(CMD_STOR, arg2 /*remote*/, arg1 /*local*/);
    }
    else {
      dataTransfer(CMD_STOR, arg1, arg1);
    }
  }
  CMD("mget") {
    REQUIRE_ARGS(1);
    mgetCommand(arg1);
  }
  CMD("mput") {
    REQUIRE_ARGS(1);
    mputCommand(arg1);
  }
  CMD("mls") {
    REQUIRE_ARGS(1);
    mlsCommand(arg1);
  }
  CMD("mdelete") {
    REQUIRE_ARGS(1);
    mdeleteCommand(arg1);
  }
  CMD("mkdir") {
    REQUIRE_ARGS(1);
    checkedRequest(CMD_MKD, arg1);
  }
  CMD("rmdir") {
    REQUIRE_ARGS(1);
    checkedRequest(CMD_RMD, arg1);
  }
  CMD("mv") {
    REQUIRE_ARGS(2);
    checkedRequest(CMD_RNFR, arg1);
    checkedRequest(CMD_RNTO, arg2);
  }
  CMD("abort") {
    checkedRequest(CMD_ABOR);
  }
  CMD("rm") {
    REQUIRE_ARGS(1);
    checkedRequest(CMD_DELE, arg1);
  }

  // -------------------- help -------------------------
  CMD("help") {
    // detailed command documentation strings are now in sftpcdoc.cpp

    NO_ARG1 {
      cout <<
        "sftpc is a command-line SafeTP client.  SafeTP is File Transfer\n"
        "Protocol (FTP), with cryptographic authentication and privacy.\n"
        "More information: http://safetp.cs.berkeley.edu\n"
        "\n"
        "Here are some of the most common commands:\n";
      printCommandList(true /*common*/);
    }

    ARG1("commands") {
      cout << "Common commands:\n";
      printCommandList(true /*common*/);
      cout << "Less-common commands:\n";
      printCommandList(false /*common*/);
    }

    ARG1("aliases") {
      cout << "Command             Alias(es)\n"
              "----------------    ---------------------";
      char const *prevExp = "";
      for (int i=0; i<numAliases; i++) {
        char const *e = aliases[i].expansion;
        char const *a = aliases[i].alias;
        if (0==strcmp(prevExp, e)) {
          // add this alias to the current line
          cout << ", " << a;
        }
        else {
          // start a new line
          prevExp = e;
          cout << endl << e;
          for (int s = strlen(e); s < 20; s++) {
            cout << " ";          // pad to aliases column
          }
          cout << a;
        }
      }
      cout << endl;
    }

    ARG1("crlf") {
      cout <<
        "Unfortunately, one of the most annoying incompatibilities\n"
        "between unix and windows is the way text file lines are\n"
        "separated.  unix separates lines with a single Line Feed (LF)\n"
        "character, whereas windows separates lines with an LF followed\n"
        "by a Carriage Return (CR) character.\n"
        "\n"
        "Thus, when transferring files between systems, it is important\n"
        "to (1) distinguish text files from binary files (the latter do\n"
        "not need translation, generally), and (2) tell the software which\n"
        "ones to translate.  As far as this note goes, you're on your own\n"
        "for (1).  (When in doubt, use binary, which is sftpc's default.)\n"
        "\n"
        "The FTP protocol handles this issue by specifying two transfer\n"
        "modes, \"ascii\" and \"image\" (binary).  In image mode, files\n"
        "are transferred without translation.  In ascii mode, files are\n"
        "treated as text, and are transferred across the network with\n"
        "CRLF separators, with each machine translating locally as necessary.\n"
        "\n"
        "To tell sftpc to transfer in ascii mode, use the \"ascii\"\n"
        "command.  To transfer in image (binary) mode, use the \"binary\"\n"
        "command.  Both commands affect all subsequent transfer operations,\n"
        "until another \"ascii\" or \"binary\" command is issued.\n"
        "\n"
        "sftpc, when working with ascii data, will report things\n"
        "that it finds suspicious.  In particular, data that is supposed\n"
        "to be CRLF-separated, but contains LFs or CRs by themselves, is\n"
        "suspicious.  Likewise, data that is supposed to be LF-separated,\n"
        "but has CRs (any at all) is suspicious.  In both circumstances,\n"
        "the number of suspicious bytes are reported, and they are\n"
        "*removed* from the data stream.\n"
        ;
    }

    else {
      // check tables
      int alias = findAlias(tok[1]);
      int cmd = findCommand(tok[1]);

      // shouldn't be documented as both
      xassert(!( alias!=-1 && cmd!=-1 ));

      if (alias != -1) {
        // alias documentation
        Alias const &a = aliases[alias];
        cout << a.alias << ": alias for \""
             << a.expansion << "\"\n";

        // show docs for expanded command
        userCommand(stringb("help " << a.expansion));
      }

      else if (cmd != -1) {
        // command documentation
        Command const &c = commands[cmd];
        cout << c.name << " " << c.args << endl
             << "  " << c.blurb << endl;

        // per-argument descs
        {
          StrtokParse tok(c.arglines, "\n");
          for (int t=0; t<tok; t++) {
            cout << "    " << tok[t] << endl;
          }
        }

        // description of command as a whole
        {
          StrtokParse tok(c.desc, "\n");
          for (int t=0; t<tok; t++) {
            cout << "  " << tok[t] << endl;
          }
        }
      }

      else {     // not found
        cout << "Unknown help topic: " << tok[1]
             << "\nTry \"help commands\"\n";
      }
    }
  }

  // --------------- last case ----------------
  else {
    // check aliases
    int i = findAlias(tok[0]);
    if (i != -1) {
      // expand the alias and recursively call this fn
      string expanded = aliases[i].expansion;
      if (tok > 1) {
        // append arguments; we only do this if there are arguments
        // because otherwise we have a trailing space, which some
        // ftpd's choke on
        expanded = stringb(aliases[i].expansion << " " << afterCmd);
      }
      userCommand(expanded);
    }

    else {
      // nothing matches
      cout << "Unknown command: " << tok[0]
           << "\nTry \"help\" or \"help commands\"\n";
    }
  }

  // manual scoping (ug)
  #undef WORDnIS
  #undef CMDIS
  #undef CMD
  #undef NO_ARG1
  #undef ARG1IS
  #undef ARG1
  #undef SET_VAR
  #undef TOGGLE_VAR
  #undef TOGGLE_ON_OFF_VAR
  #undef REQUIRE_ARGS
  #undef ELSE_BAD_ARG1
}


void SFTPC::dumpVariables()
{
  PVAL(security);
  PVAL(securityName);

  PVAL(serverPort);
  PVAL(serverName);

  PVAL(control);
  //PVAL(dataBuffer);

  PVAL(PBSZ);
  //PVAL(digt);
  PVAL(sentFirstAuth);
  PVAL(authenticated);
  cout << "data channel protection: " << getDSLString(dataSecLevel) << endl;

  PVAL(requestedPBSZ);

  PVAL(transferPassively);
  PVAL(asciiTransfers);
  PVAL(localConventionIsCRLF);

  PVAL(printOutgoing);
  PVAL(interactivePrompting);
  PVAL(quitProgram);
  PVAL(printHashMarks);

  PVAL(useRfc959);
  PVAL(acceptNewKeysSilently);
  PVAL(auto959Dropdown);

  PVAL(showDiagnostics);
  PVAL(showAdats);
  PVAL(useDebuggingLogin);
  PVAL(quitAfterNegotiation);
  PVAL(binaryTransferAnyway);
  PVAL(localGlobbing);
}


int SFTPC::emptyResponseQueue()
{
  int ret=0;

  // since we hang if there isn't a reply, we'd better check
  while (control->hasUnprocessedData() ||
         pollReadable(control->socket)) {
    // improve user-friendliness on server closure
    checkClosed(control->socket);    // may throw xSocket

    // read reply
    Reply reply = readReply();

    // DIGT and ADAT processing
    handleAdatAndDigt(reply);

    // since we don't know what this reply was for,
    // there isn't anything intelligent to do with
    // its reply code, so we'll just ignore it

    ret++;
  }

  return ret;
}



// interactive loop to login
// returns false to tell caller to exit program
bool SFTPC::login()
{
  if (useDebuggingLogin) {       // just me testing at home
    // I should mention that this 'debug' login is *not* something
    // sftpd treats differently.  Rather, it is the name of the
    // account I have defined on my debugging ftp daemons for this
    // purpose.  sftpd is oblivious to this fact.

    checkedRequest(CMD_USER, "pepe");
    checkedRequest(CMD_PASS, "pepe");    // clever password, eh?

    // if no exception thrown, we're in
    return true;
  }

  string defaultUsername = getCurrentUsername();

  // loop until successful login
  for(;;) {
    // get username
    cout << "\nUser name (Enter = " << defaultUsername << ")? ";
    cout.flush();

    string username;
    cin >> username;

    if (username == string("quit") || username == string("exit")) {
      // I sometimes do this, trying to exit 'ftp', and it thinks it's
      // a user name, and traps me in passwd prompt where ^C doesn't
      // work.. so let's just take the obvious hint and bail.
      return false;
    }

    if (username == string("")) {
      username = defaultUsername;
    }

    try {
      checkedRequest(CMD_USER, username);
    }
    catch (xReply &) {
      // this is unlikely, because servers don't typically confirm
      // whether a username is valid

      //cout << x.why() << endl;
      // we can always assume that a reply was already
      // printed, so there's no need to print it here
      continue;
    }


    // if the user name was accepted, let's change the
    // default (for this session) to what the user just
    // typed
    defaultUsername = username;


    // get password
    string password;
    if (username != string("ftp") &&
        username != string("anonymous")) {
      // read password without echoing to terminal
      char const *prompt =
        isControlChannelEncrypted()?
          "Password: " :
          "Password (will be sent as cleartext!): ";
      password = readNonechoString(prompt);
    }
    else {
      // for anonymous login, echo password, since most
      // places want an email addr as a password
      cout << "Anonymous login; email addr? ";
      cout.flush();

      cin >> password;
    }

    try {
      checkedRequest(CMD_PASS, password);
    }
    catch (xReply &) {
      //cout << x.why() << endl;
      continue;
    }


    // done
    return true;   // don't exit program
  }
}

    
// returns false to mean the server (or the current mode) cannot
// support the requested security level; other errors yield exceptions
bool SFTPC::setDataEncryption(DataSecurityLevel newLevel)
{
  if (useRfc959) {
    cout << "sftpc: I am current in 959 (no encryption) mode, so\n"
            "       data channel protection is unavailable.\n";
    return false;
  }

  // if we're turning on data encryption for the first time,
  // negotiate a PBSZ
  if (!negotiatedPBSZ()) {
    // fix: was checking that newLevel != DSL_CLEAR, but 2228
    // specifies that PBSZ must come first even for "PROT C"
    requestPBSZ();
  }

  // figure out how to communicate the desire for this protection
  // level via the control channel
  char protCode;
  if (newLevel != DSL_CLEAR) {
    xassert(newLevel & security->data().getSupportedProtLevels());
    protCode = security->data().getCodeForLevel(newLevel);
  }
  else {
    // cleartext is implemented in sftpc directly
    protCode = 'C';
  }

  // do so
  try {
    checkedRequest(CMD_PROT, stringb(protCode));
  }
  catch (xReply &) {
    cout << "The server does not support the requested data security level.\n";
    return false;
  }

  // if it succeeded, we're in the new level
  dataSecLevel = newLevel;
  return true;
}


// for extracting things like "PBSZ=number"
bool embeddedValue(int &value, char const *text, char const *prefix)
{
  char const *p = strstr(text, prefix);
  if (p != NULL) {
    value = atoi(p+strlen(prefix));
    return true;
  }
  else {
    return false;
  }
}


// general complaint about server's protocol behavior
void protocolError(char const *message)
{
  // it would be nice to have an XProtocol or something, instead
  // of just xBase...
  xfailure(message);
}


void SFTPC::requestPBSZ()
{
  // ask for user's (or sftpc default) pbsz
  sendRequest(CMD_PBSZ, stringb(requestedPBSZ));

  // loop for processing PBSZmin
  bool alreadyPBSZmin = false;     // prevent multiple pbszmin
  ReplyCode code;
  for (;;) {
    // see what server thinks about it
    Reply reply = readFinalReply();
    string text = reply.getLastText();
    code = reply.getCode();

    // ok?
    if (first(code) == RCF_POSITIVE_COMPLETION) {
      // server might have said pbsz is too big; check
      // for "PBSZ=number"
      int serverPBSZ;
      if (embeddedValue(serverPBSZ, text, "PBSZ=")) {
        // user server's pbsz
        PBSZ = serverPBSZ;
        cout << "Server adjusted PBSZ from " << requestedPBSZ
             << " to " << PBSZ << ".\n";
      }
      else {
        // use user's pbsz
        PBSZ = requestedPBSZ;
      }
      break;
    }

    // pbsz too small?
    if (code == RC_PBSZ_TOO_SMALL) {
      // prevent inf. loop from buggy server
      if (alreadyPBSZmin) {
        protocolError("server sent PBSZmin twice!");
      }
      alreadyPBSZmin = true;

      // look for "PBSZmin=number"
      int serverPBSZ;
      if (embeddedValue(serverPBSZ, text, "PBSZmin=")) {
        // change what we're requesting, and do it again
        cout << "warning: We requested PBSZ of " << requestedPBSZ
             << " but server says minimum is " << serverPBSZ << ".\n";
        if (serverPBSZ > 10000000 /*10M*/) {
          cout << "       :  That's much too big.  Bailing...\n";
          protocolError("server's PBSZmin is bigger than 10M");
        }
        else if (serverPBSZ > 1000000 /*1M*/) {
          cout << "       :  This seems extreme... possibly a bug.\n";
        }

        requestedPBSZ = serverPBSZ;
        continue;     // re-try pbsz command
      }
      else {
        protocolError(stringb(
          "Server responded with " << (int)RC_PBSZ_TOO_SMALL <<
          " but \"PBSZmin=\" is missing.\n"));
      }
    }

    // more general error
    xReply temp(reply);
    THROW(temp);
  }

  // after all this hassling, verify pbsz is sane
  xassert(PBSZ >= security->data().minimumPBSZ());

  // finally, allocate the buffer itself
  dataBuffer.setAllocated(PBSZ);
}


bool SFTPC::isControlChannelEncrypted() const
{
  return authenticated;
}

bool SFTPC::isDataChannelEncrypted() const
{
  return dataSecLevel != DSL_CLEAR;
}

bool SFTPC::isDigtActive() const
{
  return sentFirstAuth && !authenticated && !useRfc959;
}

bool SFTPC::negotiatedPBSZ() const
{
  return PBSZ > 0;
}


// main interactive loop
int SFTPC::doit()
{
  cout << "sftpc version " << SFTP_version << "\n";

  // get sockets library ready (mostly for Win32)
  socket_lib_init();

  #ifdef SIGPIPE
    // ignore broken-pipe signals
    signal(SIGPIPE, SIG_IGN);
  # endif

  if (!useRfc959) {     // skip all random seed and key stuff for 959
    // check for random seed, generate it if necessary
    if (!isRandomSeedThere()) {
      getEntropyFromConsole(1024);
      saveRandomSeed();
      if (!isRandomSeedThere()) {
	cout << "The random seed still isn't there.. ?!\n";
	return 1;
      }
    }

    // read the random seed normally, so its entropy will be
    // used in key generation (if necessary)
    readRandomSeed();

    // generate ElGamal keys if necessary; we hardcode 1024 bits here
    // for convenience; users can always run 'makekeys' manually and
    // specify a larger key size
    if (!sm_testKey("ElGamal/public.key") ||
	!sm_testKey("ElGamal/private.key")) {
      if (!GenerateElgamalKey(ElGamal_1024bit_parameters, 1024)) {
	cout << "Failed to create ElGamal keys\n";
	return 1;
      }
    }
  }

  // resolve host name
  diagnostic("resolving " << serverName << "...");
  serverAddress = resolveHostName(serverName);

  // connect to server
  string serverAddrAndPort = formatAddress(serverAddress, serverPort);
  diagnostic("connecting to " << serverAddrAndPort << "...");
  {
    SOCKET socket = connect_socket(serverAddress, serverPort);
    control = new SocketLineReader(socket);
  }

  // report connection
  cout << "Connected to " << serverName <<
          " (" << serverAddrAndPort << ").\n";

  // read the opening reply
  readAndCheckFinalReplyCode();

  // establish data and control channel protections
  {
    int retval;
    if (!negotiationPart1(retval)) {
      return retval;         // failure, typically
    }
  }

  // during the SafeTP daemon install, reaching this point is adequate
  // as a preliminary test
  if (quitAfterNegotiation) {
    return 0;    // successful negotiation
  }

  // log the user in
  if (!login()) {
    return 0;    // user asked to quit
  }

  // print some intro stuff
  cout << "Help is available: type \"help\" (no quotes, case sensitive).\n";

  // drop into interactive loop
  while (!quitProgram) {
    // debugging: verify queue is empty before printing prompt
    {
      int extras = emptyResponseQueue();
      if (extras != 0) {
        cout << "warning: " << extras << " extra replies from server!\n";
      }
    }

    // print prompt
    cout << "\nsftpc> ";
    cout.flush();

    // block until user supplies input
    string input;
    cin >> input;
    
    // did user type ^D, or is this the end of a script?
    if (cin.eof()) {
      input = "quit";      // pretend they typed "quit"
    }

    try {
      userCommand(input);
    }
    catch (xSocket &x) {
      // connection closed
      cout << "connection closed: " << x << endl;
      break;
    }
    catch (xReply &) {
      // server responds with an error; the server's error message was
      // already printed, so don't bother printing anything else
      //cout << "your input, \"" << input << "\", was not understood\n";
      diagnostic("caught xReply at command level");
    }
    catch (xBase &x) {
      // it is possible we should distinguish between errors that
      // arise from unexpected protocol behavior by the server, and
      // errors from internal failures (like failed assertions for
      // bounds checks); but for now we'll just print it and proceed
      cout << "Exception: " << x << endl;
    }
  }

  // close last socket, report leakage
  close_socket(control->socket);
  diagnostic("sftpc exiting, " << numSocketsOpen() << " sockets leaked");

  // helping to track down the cpu-chew bug
  diagnostic("gonna call deinit");
  socket_lib_deinit();
  diagnostic("deinit finished");

  if (!useRfc959) {
    saveRandomSeed();
  }
  return 0;
}


// process command line
// returns false if the program should exit
bool SFTPC::processCommandLine(int argc, char *argv[], int &retval)
{
  #define FAILRETURN(val)  retval = val; return false  // user supplies semicolon

  bool errors = false;
  int nonDashArgs = 0;
  for (int a=1; a<argc; a++) {
    bool oneError = false;
    char const *argstring = argv[a] + 2;

    if (argv[a][0] != '-') {
      argstring = argv[a];
      nonDashArgs++;
      if (nonDashArgs == 1) {      // first arg: hostname
        goto processHostname;
      }
      if (nonDashArgs == 2) {      // second arg: port number
        goto processPort;
      }
      else {                       // too many args
        oneError = true;
      }
    }
    else {
      // used letters: acdhimnpstvzAIQX9.@
      switch (argv[a][1]) {
        case 'v': {        // version
          int minv, maxv;
          knownDSAPublicKeyVersions(minv, maxv);

          printf("sftpc version %s\n"
                 "  Compiled on " __DATE__ " at " __TIME__ "\n"
                 "  Known DSA public key versions: min=%d, max=%d\n",
                 SFTP_version, minv, maxv);
          FAILRETURN(0);
        }

        case 'm':
          securityName = argstring;
          break;

        processPort:
        case 'p':          // server port
          serverPort = atoi(argstring);
          break;

        processHostname:
        case 's':          // server address
          serverName = argstring;
          break;

        case 'z':
          requestedPBSZ = atoi(argstring);
          break;

        case 'i':
          transferPassively = true;
          break;

        case 'I':
          transferPassively = false;
          break;  

        case 'd':
          showDiagnostics = true;
          break;

        case 'a':          // print ADATs
          showAdats = true;
          break;

        case '9':
          useRfc959 = true;
          dataSecLevel = DSL_CLEAR;   // -c implied by -9 now
          break;

        case 'c':
          dataSecLevel = DSL_CLEAR;
          break;                   
          
        case 't':
          dataSecLevel = DSL_INTEGRITY;
          break;

        case '.':
          useDebuggingLogin = true;
          break;

        case '@':
          // just write the html docs and exit
          writeHTMLCommands();
          FAILRETURN(0);

        case 'n':
          quitAfterNegotiation = true;
          break;

        case 'X':
          acceptNewKeysSilently = true;
          break;

        case 'Q':
          auto959Dropdown = true;
          break;

        case 'A':
          alwaysAcceptServerIPMismatch = true;
          break;

        case 'h':
          // purpose of defining this is to ensure we reserve
          // an option for printing the help text
          // (dropthru to next case)
        default:
          oneError = true;
          break;
      }
    }

    if (oneError) {
      // "-h" is treated like an error in most respects,
      // but shouldn't generate an actual error message
      if (!( argv[a][0] == '-' && argv[a][1] == 'h' )) {
        fprintf(stderr, "error in argument: %s\n", argv[a]);
      }

      // cause usage to be printed and program to exit
      errors = true;
    }
  }

  // print usage if any argument errors (or missing hostname)
  if (errors || nonDashArgs == 0) {
    fprintf(stderr,
      "usage: sftpc [options] hostname [port]\n"
      "  general options:\n"
      "    -v             print version\n"
      "    -h             print this message\n"
      "  compatibility:\n"
      "    -i             use passive-mode data transfers (this is the default)\n"
      "    -I             use active-mode data transfers (opposite of -i)\n"
      "    -9             behave as a normal RFC 959 client (no encryption)\n"
      "  debugging:\n"
      "    -d             print diagnostic (debugging) messages\n"
      "    -a             print ADATs\n"
      "  data channel protection (passwd is encrypted regardless):\n"
      "    -c             turn off data encryption\n"
      "    -t             data channel integrity-only mode\n"
      "  security check overrides:\n"
      "    -X             accept new server keys without asking\n"
      "    -Q             automatically use RFC 959 if secure connection fails\n"
      "    -A             always accept when server IP mismatch is detected\n"
      "  E.g.: sftpc york.cs.berkeley.edu\n"

      // don't print stuff that 99% of users don't care about
      //"    -m<mechanism>  use named security mechanism (default: %s)\n"
      //"    -p<port>       contact server at port N (default: %d)\n"
      //"    -s<hostname>   contact named server (default: %s)\n"
      //"    -z<pbsz>       set PBSZ (default: %ld)\n"
      //"    -n             (debugging) quit after protocol negotiation\n"
      //"    -.             (debugging) use debug login\n"
      //"    -@             (development) write html to stdout, then exit\n"
      //,
      //securityName.pcharc(),
      //serverPort,
      //serverName.pcharc(),
      //requestedPBSZ
      );
    FAILRETURN(1);
  }

  if (!showDiagnostics) {
    // turn off exception reports as well, when diagnostics are off
    xBase::logExceptions = false;
  }
  else {
    // turn on digest echoes
    //DigestComputer::echoDigestInput = true;
  }

  return true;    // success; don't exit program

  #undef FAILRETURN
}


int SFTPC::main(int argc, char *argv[])
{
  try {
    // setup error handling
    nonportFail = xSysError::xsyserror;

    // get default server port
    //serverPort = getServByName("safetp");
    // 9/2/99: after talking to Dan, it is decided that using the local
    //         machine's definition of safetp's port is not ideal.. so
    //         we'll just use 21 as the permanent default

    if (serverPort == 0) {
      serverPort = getServByName("ftp");
    }
    if (serverPort == 0) {
      // use hardcoded value if getservbyname doesn't help
      serverPort = FTPD_PORT;
    }

    // switches 'n' stuff
    {
      int retval;
      if (!processCommandLine(argc, argv, retval)) {
        return retval;
      }
    }

    // the real deal
    return doit();
  }
  catch (xBase &x) {
    cout << "Exception: " << x.why() << endl;
    return 4;
  }
}


// true OS entry
int main(int argc, char *argv[])
{
  SFTPC sftpc;
  return sftpc.main(argc, argv);
}


// ------------ KeyPolicy funcs ---------------------------
void SFTPC::keyReceived(IPAddress addr, DataBlock const &key)
{
  // check and print the brand
  DSABrandedPublicKey bkey(key);
  cout << "Public key for " << formatAddress(addr) <<
          ": " << bkey.getName() << endl;
}

void SFTPC::keyMatched(IPAddress /*addr*/)
{
  // not particularly interesting
}


void printBrandedKey(char const *ctx, DSABrandedPublicKey const &key)
{
  cout << "  " << ctx << ":\n"
       << "    Brand: " << key.getName() << endl
       << "    Sha:   " << key.getShaString() << endl;
}

SFTPC::MismatchAction SFTPC::getMismatchAction(
  IPAddress addr, DataBlock const &sentKey, DataBlock const &storedKey,
  int storedKeyCopies, IPAddress const *copyAddrs)
{
  // extract the brands; if the sentKey is bad, we'll throw xsecurity
  // and it will propagate; we assume the storedKey is not bad
  DSABrandedPublicKey brandedSentKey(sentKey);
  DSABrandedPublicKey brandedStoredKey(storedKey);

  // used twice
  string addrStr = formatAddress(addr);

  cout << "\n"
          "      ***********************\n"
          "      *    KEY MISMATCH!    *   for server at " << addrStr << ":\n"
          "      ***********************\n"
          "\n";
  printBrandedKey("Sent by server", brandedSentKey);
  printBrandedKey("Stored in database", brandedStoredKey);

  // loop over help requests
  for(;;) {
    cout << "\nOptions to proceed:\n"
            "  1. Explain the situation more fully, then ask me again.\n"
            "  5. Reject the key, and abort the connection.\n"
            "  6. Use the new key just this once; don't store it.\n"
            "  7. Replace the stored key for " << addrStr << " with the new one,\n"
            "     and use it.\n";

    // if this is a key associated with more than one server, offer the
    // user the choice of replacing them all, and list their addresses
    if (storedKeyCopies > 1) {
      cout <<
            "  8. Replace the " << storedKeyCopies << " stored keys, that match the old one,\n"
            "     with the new key, and use it.  Addresses with this key:";
      for (int i=0; i<storedKeyCopies; i++) {
        if (i%5 == 0) {   // line start
          cout << "\n       ";
        }
        cout << formatAddress(copyAddrs[i]) << " ";
      }
      cout << endl;
    }

    char ch = getOneChar("Choice? ", storedKeyCopies>1? "15678" : "1567");
    switch (ch) {
      default: xfailure("?");
      case '5':
        return MA_REJECT;

      case '6':
        cout << "Using the key for this session only, and proceeding...\n";
        return MA_ALLOW_ONCE;

      case '7':
        cout << "Replacing one key, and proceeding...\n";
        return MA_REPLACE_THIS;

      case '8':
        cout << "Replacing all matching keys, and proceeding...\n";
        return MA_REPLACE_ALL;

      case '1':
        cout << "\n"
                "Under normal circumstances, this should not happen.  There are\n"
                "two likely possibilities:\n"
                "  a) The server's administrator has created new keys.\n"
                "  b) An attacker is attempting to intercept your password.\n"
                "To distinguish these cases, you should use an alternate\n"
                "means, such as ssh, to retrieve the server's correct current\n"
                "key.  Only then, if it matches, should you accept it.\n"
                "(If it doesn't, inform the administrator!)\n"
                ;
        break;      // re-prompt
    }
  }
}


SFTPC::NewServerAction SFTPC::getNewServerAction(
  IPAddress addr, DataBlock const &sentKey)
{
  DSABrandedPublicKey brandedSentKey(sentKey);

  cout << "\nFirst contact key sent by " << formatAddress(addr) << "\n";
  printBrandedKey("Sent by server", brandedSentKey);

  if (acceptNewKeysSilently) {
    cout << "New key accepted silently as per user request.\n";
    return NSA_ACCEPT;
  }

  // count how many keys match this one
  int numMatching;
  delete[] keyDatabase.getMatchingAddrs(numMatching, sentKey);
  if (numMatching > 0) {
    // we already have this key somewhere, so let's assume this
    // key came from a multi-homed server, and accept the key
    cout << "We've seen this key before; accepting as multi-homed.\n";
    return NSA_ACCEPT;
  }

  // loop over help requests
  for (;;) {
    cout << "\nOptions to proceed:\n"
            "  1. Explain the situation more fully, then ask me again.\n"
            "  2. Reject the key, and abort the connection.\n"
            "  3. Use the new key just this once; don't store it.\n"
            "  4. Store the key for future use, and use it now.\n"
            ;

    char ch = getOneChar("Choice? ", "1234");
    switch (ch) {
      default: xfailure("?");
      case '2':
        return NSA_REJECT;

      case '3':
        cout << "Accepting key this time, and proceeding...\n";
        return NSA_ACCEPT_ONCE;

      case '4':
        cout << "Accepting key, and proceeding...\n";
        return NSA_ACCEPT;

      case '1':
        cout << "\n"
                "The database of already-seen server keys (stored in\n" <<
                userConfigDir() << ") does not contain a key for\n" <<
                formatAddress(addr) << ".\n"
                "\n"
                "There is no cryptographic reason to believe that this key\n"
                "is authentic, since this is your first contact with this\n"
                "server.  Ideally, you should confirm this key via some\n"
                "other mechanism, such as ssh or a secure web site.\n"
                "\n"
                "However, since (for example) ssh has the same vulnerability\n"
                "on the first connection, and attacks that exploit this weakness\n"
                "are (to our knowledge) rare, it's probably ok to accept the key.\n"
                ;
        break;    // re-prompt user
    }
  }
}


void SFTPC::serverIPMismatch(IPAddress ipServerReported,
                             IPAddress ipConnectedTo,
                             DataBlock const &)
{
  // no need to print the server key; it was already printed

  cout << "Server claims its IP address is "
       << formatAddress(ipServerReported)
       << ", but I think it is "
       << formatAddress(ipConnectedTo) << ".\n";

  if (alwaysAcceptServerIPMismatch) {
    // allow things to proceed, but we at least print the message
    // to at least provide lip service protection...
    return;
  }

  for (;;) {
    cout << "\nOptions to proceed:\n"
            "  e: Explain the situation more fully, then ask me again.\n"
            "  a: Accept and continue.\n"
            "  r: Reject the connection and quit.\n";

    char ch = getOneChar("Choice? ", "ear");
    switch (ch) {
      default: xfailure("?");
      case 'e':
        cout << "\n"
                "There is a possible attack where someone might redirect\n"
                "a SafeTP connection from a secure server to a less-secure\n"
                "server, in hopes of stealing passwords to the secure machine.\n"
                "\n"
                "To prevent this, the SafeTP server reports its own IP address\n"
                "in the protocol stream, and the client checks this.\n"
                "\n"
                "However, if the server is behind a firewall that does IP\n"
                "address translation, the check will also fail, even though\n"
                "there is no attack.  In this case, the fix is for the server\n"
                "administrator to add '-i" << formatAddress(ipConnectedTo) <<
                  "' to sftpd's command line.\n"
                "\n"
                "As long as the key branding above matches what you expect,\n"
                "you can safely accept the connection.\n"
                "\n"
                "If you want to always accept, use the -A flag to sftpc.\n"
                ;
        break;     // re-prompt

      case 'a':
        return;

      case 'r':
        xsecurity("Aborting connection due to server IP mismatch.");
    }
  }
}


// ----------------- EncryptedInputStreamer --------------------
int EncryptedInputStreamer::read(char *buffer, int maxlen)
{
  xassert(maxlen > 0);    // otherwise potential infinite loop

  if (curpos == sftpc.dataBuffer.getDataLen()) {
    // need to read more from the socket
    int len = sftpc.readEitherWay(dataChannel);
    totlen += len;
    if (len == 0) {
      // EOF
      return 0;
    }
    curpos = 0;
  }

  // copy some of the data buffer
  int amtToCopy = mymin(sftpc.dataBuffer.getDataLen() - curpos,
                        maxlen);
  memcpy(buffer, sftpc.dataBuffer.getDataC() + curpos, amtToCopy);

  // advance state
  curpos += amtToCopy;
  return amtToCopy;
}


// -------------------- EncryptedOutputStreamer ----------------
void EncryptedOutputStreamer::write(char const *data, int datalen)
{
  while (datalen > 0) {
    // copy as much of 'data' that will fit into current buffer
    int room = mymin(cleartextLen - curpos, datalen);
    xassert(curpos+room <= sftpc.dataBuffer.getAllocated());
    memcpy(sftpc.dataBuffer.getData()+curpos, data, room);
    curpos += room;
    data += room;
    datalen -= room;

    // send this if dataBuffer is full
    if (curpos >= cleartextLen) {
      xassert(curpos == cleartextLen);     // defensive programming
      sendIt();
    }
  }
}


void EncryptedOutputStreamer::sendIt()
{
  sftpc.dataBuffer.setDataLen(curpos);
  sftpc.writeEitherWay(dataChannel);
  totlen += curpos;
  curpos = 0;
}


EncryptedOutputStreamer::~EncryptedOutputStreamer()
{
  // empty the buffer
  if (curpos > 0) {
    sendIt();
  }

  // send a final, empty block (does a no-op if not encrypted)
  sendIt();    // curpos is 0, so this will be 0-length
}


// --------------- EchoStream ----------------------
void EchoStream::saw(int numChars)
{
  outstanding += numChars;
  while (outstanding >= echoFreq) {
    cout << echoChar;
    outstanding -= echoFreq;
  }
}


// ----------------- EchoInputStream ---------------------------
int EchoInputStream::read(char *buffer, int buflen)
{
  int len = input.read(buffer, buflen);
  saw(len);
  return len;
}


// ---------------- EchoOutputStream ---------------------
void EchoOutputStream::write(char const *data, int len)
{
  output.write(data, len);
  saw(len);
}


// ----------------- trash --------------------------
#if 0
  // I can't understand why my Linux ftpd is pissed off, so
  // let's see if it wants me to eat the 150
  if (req.getCmd() == CMD_NLST) {
    diagnostic("waiting for the 150");
    Reply interReply = readReply();
    if (interReply.getCode() == RC_FILE_STATUS_OK) {
      diagnostic("got the 150 as expected");
    }
    else {
      // most likely we won't even get here if a 150 isn't waiting,
      // since the 226 certainly won't have been sent yet...
      diagnostic("what the hell?  that was supposed to be a 150");
    }
  }

YY_INPUT(buf,result,max_size)


      // process each incoming block of data
      for(;;) {
        int len = readEitherWay(dataChannel);
        if (len == 0) {
          break;
        }

        char *lineStart = dataBuffer.getData();
        char *bufferEnd = lineStart + dataBuffer.getDataLen();
        char *p = lineStart;

        // look for CRLF
        while (p < bufferEnd-1 && !(p[0] == '\r' && p[1] == '\n')) {
          p++;
        }

        if (p < bufferEnd-1) {    // found CRLF
          // change the existing line ending
          p[0] = '\n';
          p++;

          // write the line, with new line ending
          if (fwrite(lineStart, 1, p-lineStart, dest) != p-lineStart) {
            // this is a lousy way to report errors... we don't have
            // the file name, it's an uninformative 'xBase', etc.
            THROW(xBase("Error writing to output file"));
          }

          // move to next line
          p = lineStart = p+1;
        }

        else {                    // reached end of buffer
          if (p[0] == '\r') {
            // possibly a CRLF pair split across buffer boundary
            pendingCR = true;
          }
          else {
            // can't be CRLF pair split, so increment p so we will
            // output this last character also
            p++;
            pendingCR = false;
          }

          // output what's at the end of the buffer










      // parse the incoming data into lines, mindful of the
      // fact that we may be decrypting data coming across
      // as blocks which may be broken across CRLF pairs
      int bytes;
      SFTPC_InputSource inputSource(this, dataChannel, bytes);
      StreamLineReader lineReader(inputSource);

      // read lines until socket closed
      string line;
      while (lineReader.getNextLine(line)) {
        // write it to the file, with only an LF at the end
        fprintf(dest, "%s\n", line.pcharc());
      }

      // note: if the file ends without a CRLF, this will append
      //       an LF anyway




void SFTPC::retrCommand(char const *fname)
{
  // open a local file to store the information
  FILE *fp = fopen(fname, "wb");
  if (!fp) {
    THROW(xBase(stringb("failed to open " << req.getData())));
  }
  FileOutputDest destStream(fp);

  // PORT, etc.
  SOCKET dataChannel = startDataTransfer(CMD_RETR, fname);

  // copy data from the socket to the file until the
  // socket is closed
  int totlen = copyFromSocketToStream(dataChannel, destStream,
                                      transferModeForCmd(CMD_RETR));

  diagnostic("received " << totlen <<
             " bytes to local file " << fname);

  fclose(fp);
  close_socket(dataChannel);
}


void SFTPC::storCommand(char const *fname)
{
  // open the local file containing the data to send
  FILE *fp = fopen(fname, "rb");
  if (!fp) {
    THROW(xBase(stringb("failed to open " << req.getData())));
  }
  FileOutputDest destStream(fp);

  // PORT, etc.
  SOCKET dataChannel = startDataTransfer(cmd, fname);

  // copy data from the file to the socket
  int totlen = copyFromStreamToSocket(destStream, dataChannel,
                                      transferModeForCmd(CMD_RETR));

  diagnostic("sent " << totlen <<
             " bytes from local file " << fname);

  fclose(fp);
  close_socket(dataChannel);
}


void SFTPC::listCommand(RequestCommand cmd, char const *arg)
{
  // PORT, etc.
  SOCKET dataChannel = startDataTransfer(cmd, fname);

  diagnostic("waiting for directory listing");

  // copy data from the socket to the terminal until the
  // socket is closed
  FileOutputStream destStream(stdout);
  copyFromSocketToFile(dataChannel, destStream,
                       transferModeForCmd(CMD_RETR));
  close_socket(dataChannel);
}

// there may be more efficient ways to do this, but who cares
static bool exists(char const *fname)
{
  FILE *fp = fopen(fname, "rb");
  if (!fp) {
    return false;
  }
  else {
    fclose(fp);
    return true;
  }
}


// ptr that deletes referrent when ptr is destroyed
// (this is a rather limited version, implemented only for its uses
// in sftpc.cpp)
template <class T>
class autoDelPtr {
  T *ptr;

public:
  autoDelPtr(T *p = NULL) : ptr(p) {}
  ~autoDelPtr() {
    if (ptr != NULL) {
      delete ptr;
    }
  }

  operator T* () { return ptr; }
  autoDelPtr& operator= (T *p) { ptr=p; return *this; }
};

#endif // 0

